#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import pathlib
from enum import IntEnum
from io import BytesIO
from itertools import chain
from typing import Any
from xml.etree import ElementTree

from packaging.version import Version

import odxtools_porsche.uds as uds
from odxtools.additionalaudience import AdditionalAudience
from odxtools.admindata import AdminData
from odxtools.audience import Audience
from odxtools.companydata import CompanyData
from odxtools.companydocinfo import CompanyDocInfo
from odxtools.companyspecificinfo import CompanySpecificInfo
from odxtools.comparaminstance import ComparamInstance
from odxtools.comparamspec import ComparamSpec
from odxtools.comparamsubset import ComparamSubset
from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compuconst import CompuConst
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compumethod import CompuMethod
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.compumethods.limit import Limit
from odxtools.compumethods.texttablecompumethod import TexttableCompuMethod
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.description import Description
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diaglayercontainer import DiagLayerContainer
from odxtools.diaglayers.basevariant import BaseVariant
from odxtools.diaglayers.basevariantraw import BaseVariantRaw
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.diaglayers.protocol import Protocol
from odxtools.diaglayers.protocolraw import ProtocolRaw
from odxtools.diagservice import DiagService
from odxtools.docrevision import DocRevision
from odxtools.exceptions import odxrequire
from odxtools.functionalclass import FunctionalClass
from odxtools.modification import Modification
from odxtools.nameditemlist import NamedItemList
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.matchingrequestparameter import MatchingRequestParameter
from odxtools.parameters.nrcconstparameter import NrcConstParameter
from odxtools.parameters.tablekeyparameter import TableKeyParameter
from odxtools.parameters.tablestructparameter import TableStructParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.parentref import ParentRef
from odxtools.physicaldimension import PhysicalDimension
from odxtools.physicaltype import PhysicalType
from odxtools.preconditionstateref import PreConditionStateRef
from odxtools.progcode import ProgCode
from odxtools.relateddoc import RelatedDoc
from odxtools.request import Request
from odxtools.response import Response, ResponseType
from odxtools.singleecujob import SingleEcuJob
from odxtools.standardlengthtype import StandardLengthType
from odxtools.state import State
from odxtools.statechart import StateChart
from odxtools.statetransition import StateTransition
from odxtools.statetransitionref import StateTransitionRef
from odxtools.structure import Structure
from odxtools.table import Table
from odxtools.tablerow import TableRow
from odxtools.teammember import TeamMember
from odxtools.unit import Unit
from odxtools.unitgroup import UnitGroup
from odxtools.unitgroupcategory import UnitGroupCategory
from odxtools.unitspec import UnitSpec
from odxtools.xdoc import XDoc

ODX_VERSION = Version("2.2.0")


class SomersaultSID(IntEnum):
    """The Somersault-ECU specific service IDs.

    These are ECU specific service IDs allocated by the UDS standard.

    """

    ForwardFlip = 0xBA
    BackwardFlip = 0xBB
    Headstand = 0xBC
    ForwardFlipCondition = 0xBD


# extend the Somersault SIDs by the UDS ones
tmp = [(i.name, i.value) for i in chain(uds.SID, SomersaultSID)]
SID: Any = IntEnum("SID", tmp)  # type: ignore[misc]

dlc_short_name = "somersault"

# document fragment for everything except the communication parameters
doc_frags = [OdxDocFragment(dlc_short_name, DocType.CONTAINER)]

# document fragments for communication parameters
cp_dwcan_doc_frags = [OdxDocFragment("ISO_11898_2_DWCAN", DocType.COMPARAM_SUBSET)]
cp_iso15765_2_doc_frags = [OdxDocFragment("ISO_15765_2", DocType.COMPARAM_SUBSET)]
cp_iso15765_3_doc_frags = [OdxDocFragment("ISO_15765_3", DocType.COMPARAM_SUBSET)]

##################
# Base variant of Somersault ECU
##################

# company datas
somersault_team_members = {
    "doggy":
        TeamMember(
            odx_id=OdxLinkId("TM.Doggy", doc_frags),
            short_name="Doggy",
            long_name="Doggy the dog",
            description=Description.from_string("<p>Dog is man's best friend</p>"),
            roles=["gymnast", "tracker"],
            department="sniffers",
            address="Some road",
            zipcode="12345",
            city="New Dogsville",
            phone="+0 1234/5678-9",
            fax="+0 1234/5678-0",
            email="info@suncus.com",
        ),
    "horsey":
        TeamMember(
            odx_id=OdxLinkId("TM.Horsey", doc_frags),
            short_name="Horsey",
            long_name="Horsey the horse",
            description=Description.from_string("<p>Trustworthy worker</p>"),
            roles=["gymnast"],
            department="haulers",
            address="Some road",
            zipcode="12345",
            city="New Dogsville",
            phone="+0 1234/5678-91",
            fax="+0 1234/5678-0",
            email="info@suncus.com",
        ),
    "slothy":
        TeamMember(
            odx_id=OdxLinkId("TM.Slothy", doc_frags),
            short_name="Slothy",
        ),
}

somersault_company_datas = {
    "suncus":
        CompanyData(
            odx_id=OdxLinkId("CD.Suncus", doc_frags),
            short_name="Suncus",
            long_name="Circus of the sun",
            description=Description.from_string("<p>Prestigious group of performers</p>"),
            roles=["circus", "gym"],
            team_members=NamedItemList([
                somersault_team_members["doggy"],
                somersault_team_members["horsey"],
            ]),
            company_specific_info=CompanySpecificInfo(related_docs=[
                RelatedDoc(
                    description=Description.from_string("<p>We are the best!</p>"),
                    xdoc=XDoc(
                        short_name="best",
                        long_name="suncus is the best",
                        description=Description.from_string("<p>great propaganda...</p>"),
                        number="1",
                        state="published",
                        date="2015-01-15T20:15:20+05:00",
                        publisher="Suncus Publishing",
                        url="https://suncus-is-the-best.com",
                        position="first!",
                    ),
                ),
            ]),
        ),
    "acme":
        CompanyData(
            odx_id=OdxLinkId("CD.ACME", doc_frags),
            short_name="ACME_Corporation",
            team_members=NamedItemList([
                somersault_team_members["slothy"],
            ]),
        ),
}

somersault_admin_data = AdminData(
    language="en-US",
    company_doc_infos=[
        CompanyDocInfo(
            company_data_ref=OdxLinkRef("CD.Suncus", doc_frags),
            team_member_ref=OdxLinkRef("TM.Doggy", doc_frags),
            doc_label="A really meaningful label",
        ),
    ],
    doc_revisions=[
        DocRevision(
            team_member_ref=OdxLinkRef("TM.Doggy", doc_frags),
            revision_label="1.0",
            state="draft",
            date="1926-07-18T11:11:11+01:00",
            tool="odxtools 0.0.1",
            modifications=[
                Modification(change="add somersault ECU", reason="we needed a new artist"),
                Modification(
                    change="increase robustness to dizzyness", reason="No alcohol anymore"),
            ],
        ),
        DocRevision(
            team_member_ref=OdxLinkRef("TM.Horsey", doc_frags),
            revision_label="1.1",
            state="released",
            date="2020-08-19T12:12:12+08:00",
            tool="odxtools 0.1",
            modifications=[
                Modification(
                    change="rename somersault ECU to somersault_assiduous to enable slothy to add somersault_lazy",
                ),
            ],
        ),
        DocRevision(
            team_member_ref=OdxLinkRef("TM.Slothy", doc_frags),
            revision_label="1.0.3.2.1.5.6",
            date="1900-01-01T00:00:00+00:00",
        ),
    ],
)

# functional classes
somersault_functional_classes = {
    "flip":
        FunctionalClass(
            odx_id=OdxLinkId("somersault.FNC.flip", doc_frags),
            short_name="flip",
            long_name="Flip",
        ),
    "session":
        FunctionalClass(
            odx_id=OdxLinkId("somersault.FNC.session", doc_frags),
            short_name="session",
            long_name="Session",
        ),
}

# additional audiences
somersault_additional_audiences = {
    "attentive_admirer":
        AdditionalAudience(
            odx_id=OdxLinkId("somersault.AA.attentive_admirer", doc_frags),
            short_name="attentive_admirer",
            long_name="Attentive Admirer",
        ),
    "anyone":
        AdditionalAudience(
            odx_id=OdxLinkId("somersault.AA.anyone", doc_frags),
            short_name="anyone",
            long_name="Anyone",
        ),
}

# diag coded types
somersault_diagcodedtypes = {
    "flag": StandardLengthType(
        base_data_type=DataType.A_UINT32,
        bit_length=1,
    ),
    "int8": StandardLengthType(
        base_data_type=DataType.A_INT32,
        bit_length=8,
    ),
    "uint8": StandardLengthType(
        base_data_type=DataType.A_UINT32,
        bit_length=8,
    ),
    "uint16": StandardLengthType(
        base_data_type=DataType.A_UINT32,
        bit_length=16,
    ),
    "float32": StandardLengthType(
        base_data_type=DataType.A_FLOAT32,
        bit_length=32,
    ),
}

somersault_physical_dimensions = {
    "time":
        PhysicalDimension(
            odx_id=OdxLinkId("somersault.PD.time", doc_frags),
            short_name="time",
            long_name="Time",
            time_exp=1,
        ),
    "temperature":
        PhysicalDimension(
            odx_id=OdxLinkId("somersault.PD.temperature", doc_frags),
            short_name="temperature",
            long_name="Temperature",
            temperature_exp=1,
        )
}

somersault_units = {
    "second":
        Unit(
            odx_id=OdxLinkId("somersault.unit.second", doc_frags),
            short_name="second",
            display_name="s",
            long_name="Second",
            description=Description.from_string("<p>SI unit for the time</p>"),
            factor_si_to_unit=1,
            offset_si_to_unit=0,
            physical_dimension_ref=OdxLinkRef.from_id(
                somersault_physical_dimensions["time"].odx_id),
        ),
    "minute":
        Unit(
            odx_id=OdxLinkId("somersault.unit.minute", doc_frags),
            short_name="minute",
            display_name="min",
            long_name="Minute",
            factor_si_to_unit=60,
            offset_si_to_unit=0,
            physical_dimension_ref=OdxLinkRef.from_id(
                somersault_physical_dimensions["time"].odx_id),
        ),
    "celsius":
        Unit(
            odx_id=OdxLinkId("somersault.unit.celsius", doc_frags),
            short_name="celsius",
            display_name="°C",
            long_name="Degrees Celcius",
            factor_si_to_unit=1,
            offset_si_to_unit=-273.15,
            physical_dimension_ref=OdxLinkRef.from_id(
                somersault_physical_dimensions["temperature"].odx_id),
        ),
}

somersault_unit_groups = {
    "european_duration":
        UnitGroup(
            short_name="european_duration",
            category=UnitGroupCategory.COUNTRY,
            unit_refs=[
                OdxLinkRef.from_id(somersault_units["second"].odx_id),
                OdxLinkRef.from_id(somersault_units["minute"].odx_id),
            ],
            long_name="Duration",
            description=Description.from_string("<p>Units for measuring a duration</p>"),
        ),
}

# computation methods
somersault_compumethods: dict[str, CompuMethod] = {
    "int_passthrough":
        IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32),
    "uint_passthrough":
        IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32),
    "float_passthrough":
        IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32),
    "boolean":
        TexttableCompuMethod(
            category=CompuCategory.TEXTTABLE,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_const=CompuConst(vt="false", data_type=DataType.A_UTF8STRING),
                    lower_limit=Limit(value_raw="0", value_type=DataType.A_UINT32),
                    upper_limit=Limit(value_raw="0", value_type=DataType.A_UINT32),
                    domain_type=DataType.A_UINT32,
                    range_type=DataType.A_UNICODE2STRING),
                CompuScale(
                    compu_const=CompuConst(vt="true", data_type=DataType.A_UTF8STRING),
                    lower_limit=Limit(value_raw="1", value_type=DataType.A_UINT32),
                    upper_limit=Limit(value_raw="1", value_type=DataType.A_UINT32),
                    domain_type=DataType.A_UINT32,
                    range_type=DataType.A_UNICODE2STRING),
            ]),
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UNICODE2STRING,
        ),
}

# data object properties
somersault_dops = {
    "num_flips":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.num_flips", doc_frags),
            short_name="num_flips",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "soberness_check":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.soberness_check", doc_frags),
            short_name="soberness_check",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "dizzyness_level":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.dizzyness_level", doc_frags),
            short_name="dizzyness_level",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "happiness_level":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.happiness_level", doc_frags),
            short_name="happiness_level",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "duration":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.duration", doc_frags),
            short_name="duration",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
            unit_ref=OdxLinkRef.from_id(somersault_units["second"].odx_id),
        ),
    "temperature":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.temperature", doc_frags),
            short_name="temperature",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
            unit_ref=OdxLinkRef.from_id(somersault_units["celsius"].odx_id),
        ),
    "error_code":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.error_code", doc_frags),
            short_name="error_code",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "boolean":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.boolean", doc_frags),
            short_name="boolean",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UNICODE2STRING),
            compu_method=somersault_compumethods["boolean"],
        ),
    "uint8":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.uint8", doc_frags),
            short_name="uint8",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "uint16":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.uint16", doc_frags),
            short_name="uint16",
            diag_coded_type=somersault_diagcodedtypes["uint16"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "float":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.float", doc_frags),
            short_name="float",
            diag_coded_type=somersault_diagcodedtypes["float32"],
            physical_type=PhysicalType(base_data_type=DataType.A_FLOAT32),
            compu_method=somersault_compumethods["float_passthrough"],
        ),
    "schroedinger_base":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.schroedinger_base", doc_frags),
            short_name="schroedinger_dop",
            diag_coded_type=somersault_diagcodedtypes["int8"],
            physical_type=PhysicalType(base_data_type=DataType.A_INT32),
            compu_method=somersault_compumethods["int_passthrough"],
        ),
    "schroedinger_lazy":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.schroedinger_lazy", doc_frags),
            short_name="schroedinger_dop",
            diag_coded_type=somersault_diagcodedtypes["uint8"],
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=somersault_compumethods["uint_passthrough"],
        ),
    "schroedinger_assiduous":
        DataObjectProperty(
            odx_id=OdxLinkId("somersault.DOP.schroedinger_assiduous", doc_frags),
            short_name="schroedinger_dop",
            diag_coded_type=somersault_diagcodedtypes["float32"],
            physical_type=PhysicalType(base_data_type=DataType.A_FLOAT32),
            compu_method=somersault_compumethods["float_passthrough"],
        ),
}

last_flip_details_table_id = OdxLinkId("somersault.table.last_flip_details", doc_frags)
last_flip_details_table_ref = OdxLinkRef.from_id(last_flip_details_table_id)

# positive responses
somersault_positive_responses = {
    "session":
        Response(
            odx_id=OdxLinkId("somersault.PR.session_start", doc_frags),
            short_name="session",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        uds.positive_response_id(
                            SID.DiagnosticSessionControl.value)),  # type: ignore[attr-defined]
                ),
                ValueParameter(
                    short_name="can_do_backward_flips",
                    byte_position=1,
                    dop_ref=OdxLinkRef("somersault.DOP.boolean", doc_frags),
                ),
            ]),
        ),
    "tester_ok":
        Response(
            odx_id=OdxLinkId("somersault.PR.tester_present", doc_frags),
            short_name="tester_present",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.TesterPresent.value)),  # type: ignore[attr-defined]
                ),
                ValueParameter(
                    short_name="status",
                    dop_ref=OdxLinkRef("somersault.DOP.uint8", doc_frags),
                    physical_default_value_raw="0",
                    byte_position=1,
                ),
            ]),
        ),
    "forward_flips_grudgingly_done":
        Response(
            odx_id=OdxLinkId("somersault.PR.grudging_forward", doc_frags),
            short_name="grudging_forward",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.ForwardFlip.value)),  # type: ignore[attr-defined]
                ),
                # TODO (?): non-byte aligned MatchingRequestParameters
                MatchingRequestParameter(
                    short_name="num_flips_done",
                    request_byte_position=2,
                    byte_position=1,
                    byte_length=1,
                ),
                ValueParameter(
                    short_name="sault_time",
                    physical_default_value_raw="255",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.duration", doc_frags),
                ),
            ]),
        ),
    "forward_flips_happily_done":
        Response(
            odx_id=OdxLinkId("somersault.PR.happy_forward", doc_frags),
            short_name="happy_forward",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.ForwardFlip.value)),  # type: ignore[attr-defined]
                ),
                # TODO (?): non-byte aligned MatchingRequestParameters
                MatchingRequestParameter(
                    short_name="num_flips_done",
                    request_byte_position=0,
                    byte_position=1,
                    byte_length=1,
                ),
                ValueParameter(
                    short_name="yeha_level",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.num_flips", doc_frags),
                ),
            ]),
        ),
    "backward_flips_grudgingly_done":
        Response(
            odx_id=OdxLinkId("somersault.PR.grudging_backward", doc_frags),
            short_name="grudging_backward",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.BackwardFlip.value)),  # type: ignore[attr-defined]
                ),
                ValueParameter(
                    short_name="num_flips_done",
                    dop_ref=OdxLinkRef("somersault.DOP.uint8", doc_frags),
                ),
                ValueParameter(
                    short_name="grumpiness_level",
                    dop_ref=OdxLinkRef("somersault.DOP.uint8", doc_frags),
                ),
            ]),
        ),
    # Note that there is no such thing as a "backwards flip done happily"!
    "status_report":
        Response(
            odx_id=OdxLinkId("somersault.PR.status_report", doc_frags),
            short_name="status_report",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.ReadDataByIdentifier.value)),  # type: ignore[attr-defined]
                ),
                ValueParameter(
                    short_name="dizzyness_level",
                    byte_position=1,
                    dop_ref=OdxLinkRef("somersault.DOP.dizzyness_level", doc_frags),
                ),
                ValueParameter(
                    short_name="happiness_level",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.happiness_level", doc_frags),
                ),
                TableKeyParameter(
                    odx_id=OdxLinkId("somersault.PR.report_status.last_pos_response_key",
                                     doc_frags),
                    short_name="last_pos_response_key",
                    table_ref=last_flip_details_table_ref,
                    byte_position=3,
                ),
                TableStructParameter(
                    short_name="last_pos_response",
                    table_key_ref=OdxLinkRef("somersault.PR.report_status.last_pos_response_key",
                                             doc_frags),
                ),
            ]),
        ),
    "set_operation_params":
        Response(
            odx_id=OdxLinkId("somersault.PR.set_operation_params", doc_frags),
            short_name="set_operation_params",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.positive_response_id(
                        SID.ForwardFlipCondition.value)),  # type: ignore[attr-defined]
                ),
            ]),
        ),
}

somersault_structures = {
    "forward_flips_grudgingly_done":
        Structure(
            short_name="forward_flips_grudgingly_done_recall",
            odx_id=OdxLinkId("somersault.struct.recall.forward_flips_grudgingly_done", doc_frags),
            parameters=somersault_positive_responses["forward_flips_grudgingly_done"].parameters),
    "forward_flips_happily_done":
        Structure(
            short_name="forward_flips_happily_done_recall",
            odx_id=OdxLinkId("somersault.struct.recall.forward_flips_happily_done", doc_frags),
            parameters=somersault_positive_responses["forward_flips_happily_done"].parameters),
    "backward_flips_grudgingly_done":
        Structure(
            short_name="backward_flips_grudgingly_done_recall",
            odx_id=OdxLinkId("somersault.struct.recall.backward_flips_grudgingly_done", doc_frags),
            parameters=somersault_positive_responses["backward_flips_grudgingly_done"].parameters),
}

# this is a hack to get around a catch-22: we need to specify the
# value of a positive response to the tester present parameter to
# specify ISO_15765_3.CP_TesterPresentMessage communication parameter,
# but we need the comparam for the raw diaglayer which we need for
# retrieving the DOP of the "status" parameter in order to convert the
# raw physical default value.
param = somersault_positive_responses["tester_ok"].parameters.status
assert isinstance(param, ValueParameter)
param._dop = somersault_dops["uint8"]
param._physical_default_value = int(odxrequire(param.physical_default_value_raw))

# negative responses
somersault_negative_responses = {
    "general":
        Response(
            odx_id=OdxLinkId("somersault.NR.general_negative_response", doc_frags),
            short_name="general_negative_response",
            response_type=ResponseType.NEGATIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.NegativeResponseId),
                ),
                MatchingRequestParameter(
                    short_name="rq_sid",
                    request_byte_position=0,
                    byte_position=1,
                    byte_length=1,
                ),
                ValueParameter(
                    short_name="response_code",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.error_code", doc_frags),
                ),
            ]),
        ),
    # the tester present request needs separate negative and positive
    # responses because it must be fully specified a-priory to be able
    # to extract it for the COMPARAMS.
    "tester_nok":
        Response(
            odx_id=OdxLinkId("somersault.NR.tester_nok", doc_frags),
            short_name="tester_nok",
            response_type=ResponseType.NEGATIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.NegativeResponseId),
                ),
                CodedConstParameter(
                    short_name="rq_sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(uds.SID.TesterPresent.value),  # type: ignore[attr-defined]
                ),
            ]),
        ),
    "flips_not_done":
        Response(
            odx_id=OdxLinkId("somersault.NR.flips_not_done", doc_frags),
            short_name="flips_not_done",
            response_type=ResponseType.NEGATIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.NegativeResponseId),
                ),
                MatchingRequestParameter(
                    short_name="rq_sid",
                    request_byte_position=0,
                    byte_position=1,
                    byte_length=1,
                ),
                NrcConstParameter(
                    short_name="reason",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=2,
                    coded_values_raw=[str(0), str(1), str(2)],
                    # possible values (TODO: make this an enum parameter):
                    # 0 -> not sober
                    # 1 -> too dizzy
                    # 2 -> stumbled
                ),
                ValueParameter(
                    short_name="flips_successfully_done",
                    dop_ref=OdxLinkRef("somersault.DOP.num_flips", doc_frags),
                    byte_position=3,
                ),
            ]),
        ),
}

somersault_global_negative_responses = {
    "general":
        Response(
            odx_id=OdxLinkId("GNR.too_hot", doc_frags),
            short_name="too_hot",
            response_type=ResponseType.GLOBAL_NEGATIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(uds.NegativeResponseId),
                ),
                CodedConstParameter(
                    short_name="too_hot_dummy",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(0xfe),
                ),
                CodedConstParameter(
                    short_name="too_hot_id",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=2,
                    coded_value_raw=str(0xa7),
                ),
                ValueParameter(
                    short_name="temperature",
                    byte_position=3,
                    dop_ref=OdxLinkRef("somersault.DOP.temperature", doc_frags),
                ),
            ]),
        )
}

# tables
flip_quality_table_id = OdxLinkId("somersault.table.flip_quality", doc_frags)
flip_quality_table_ref = OdxLinkRef.from_id(flip_quality_table_id)
somersault_tables = {
    "last_flip_details":
        Table(
            odx_id=last_flip_details_table_id,
            short_name="last_flip_details",
            long_name="Flip Details",
            description=Description.from_string(
                "<p>The details the last successfully executed request</p>"),
            semantic="DETAILS",
            key_label="key",
            struct_label="response",
            key_dop_ref=OdxLinkRef.from_id(somersault_dops["uint8"].odx_id),
            table_rows_raw=[
                TableRow(
                    table_ref=last_flip_details_table_ref,
                    odx_id=OdxLinkId("somersault.table.last_flip_details.none", doc_frags),
                    short_name="none",
                    long_name="No Flips Done Yet",
                    key_raw="0",
                    description=Description.from_string("<p>We have not done any flips yet!</p>"),
                    semantic="DETAILS-KEY",
                    dop_ref=OdxLinkRef.from_id(somersault_dops["soberness_check"].odx_id),
                ),
                TableRow(
                    table_ref=last_flip_details_table_ref,
                    odx_id=OdxLinkId("somersault.table.last_flip_details.forward_grudging",
                                     doc_frags),
                    short_name="forward_grudging",
                    long_name="Forward Flips Grudgingly Done",
                    key_raw="3",
                    structure_ref=OdxLinkRef.from_id(
                        somersault_structures["forward_flips_grudgingly_done"].odx_id),
                    description=Description.from_string(
                        "<p>The the last forward flip was grudgingly done</p>"),
                    semantic="DETAILS-KEY",
                ),
                TableRow(
                    table_ref=last_flip_details_table_ref,
                    odx_id=OdxLinkId("somersault.table.last_flip_details.forward_happy", doc_frags),
                    short_name="forward_happily",
                    long_name="Forward Flips Happily Done",
                    key_raw="5",
                    structure_ref=OdxLinkRef.from_id(
                        somersault_structures["forward_flips_happily_done"].odx_id),
                ),
                TableRow(
                    table_ref=last_flip_details_table_ref,
                    odx_id=OdxLinkId("somersault.table.last_flip_details.backward", doc_frags),
                    short_name="backward_grudging",
                    long_name="Backward Flips",
                    key_raw="10",
                    structure_ref=OdxLinkRef.from_id(
                        somersault_structures["backward_flips_grudgingly_done"].odx_id),
                ),
            ],
        )
}

# requests
somersault_requests = {
    "start_session":
        Request(
            odx_id=OdxLinkId("somersault.RQ.start_session", doc_frags),
            short_name="start_session",
            long_name="Start the diagnostic session & do some mischief",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.DiagnosticSessionControl.value),  # type: ignore[attr-defined]
                ),
                CodedConstParameter(
                    short_name="id",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(0x0),
                ),
                ValueParameter(
                    short_name="bribe",
                    physical_default_value_raw="0",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.uint8", doc_frags),
                ),
            ]),
        ),
    "stop_session":
        Request(
            odx_id=OdxLinkId("somersault.RQ.stop_session", doc_frags),
            short_name="stop_session",
            long_name="Terminate the current diagnostic session",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.DiagnosticSessionControl  # type: ignore[attr-defined,arg-type]
                        .value),
                ),
                CodedConstParameter(
                    short_name="id",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(0x1),
                ),
            ],
        ),
    "tester_present":
        Request(
            odx_id=OdxLinkId("somersault.RQ.tester_present", doc_frags),
            short_name="tester_present",
            long_name="Prevent the current diagnostic session from timing out",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.TesterPresent.value),  # type: ignore[attr-defined,arg-type]
                ),
                CodedConstParameter(
                    short_name="id",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(0x0),
                ),
            ],
        ),
    "set_operation_params":
        Request(
            odx_id=OdxLinkId("somersault.RQ.set_operation_params", doc_frags),
            short_name="set_operation_params",
            long_name="Specify the mode of operation for the ECU; e.g. if rings "
            "of fire ought to be used for maximum effect",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.ForwardFlipCondition  # type: ignore[attr-defined,arg-type]
                        .value),
                ),
                ValueParameter(
                    short_name="use_fire_ring",
                    byte_position=1,
                    dop_ref=OdxLinkRef("somersault.DOP.boolean", doc_frags),
                ),
            ],
        ),
    "forward_flips":
        Request(
            odx_id=OdxLinkId("somersault.RQ.do_forward_flips", doc_frags),
            short_name="do_forward_flips",
            long_name="Do forward somersaults & some other mischief",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.ForwardFlip.value),  # type: ignore[attr-defined,arg-type]
                ),
                ValueParameter(
                    short_name="forward_soberness_check",
                    dop_ref=OdxLinkRef("somersault.DOP.soberness_check", doc_frags),
                    byte_position=1,
                    # value must be 0x12 for the request to be accepted
                ),
                ValueParameter(
                    short_name="num_flips",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.num_flips", doc_frags),
                ),
            ],
        ),
    "backward_flips":
        Request(
            odx_id=OdxLinkId("somersault.RQ.do_backward_flips", doc_frags),
            short_name="do_backward_flips",
            long_name="Do a backward somersault & some other mischief",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.BackwardFlip.value),  # type: ignore[attr-defined,arg-type]
                ),
                ValueParameter(
                    short_name="backward_soberness_check",
                    dop_ref=OdxLinkRef("somersault.DOP.soberness_check", doc_frags),
                    byte_position=1,
                    # value must be 0x21 for the request to be accepted
                ),
                ValueParameter(
                    short_name="num_flips",
                    byte_position=2,
                    dop_ref=OdxLinkRef("somersault.DOP.num_flips", doc_frags),
                ),
            ],
        ),
    "report_status":
        Request(
            odx_id=OdxLinkId("somersault.RQ.report_status", doc_frags),
            short_name="report_status",
            long_name="Report back the current level of dizzy- & happiness.",
            parameters=[
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(
                        SID.ReadDataByIdentifier  # type: ignore[attr-defined,arg-type]
                        .value),
                ),
                CodedConstParameter(
                    short_name="id",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=1,
                    coded_value_raw=str(0x0),
                ),
            ],
        ),
    "schroedinger":
        Request(
            odx_id=OdxLinkId("somersault.RQ.schroedinger", doc_frags),
            short_name="schroedinger_request",
            parameters=NamedItemList([
                ValueParameter(
                    short_name="schroedinger_param",
                    long_name="Parameter where the DOP changes dending on how you "
                    "look at the SNREF to it",
                    byte_position=0,
                    dop_snref="schroedinger_dop",
                ),
            ]),
        ),
}

# services
somersault_services = {
    "start_session":
        DiagService(
            odx_id=OdxLinkId("somersault.service.session_start", doc_frags),
            short_name="session_start",
            pre_condition_state_refs=[
                PreConditionStateRef(
                    ref_id="charts.annoyed.states.in_bed",
                    ref_docs=doc_frags,
                ),
                # note that the standard does not allow to specify
                # relations other than equivalence for the specified
                # value (larger-than would be more appropriate here...)
                PreConditionStateRef(
                    ref_id="charts.angry.states.in_bed",
                    ref_docs=doc_frags,
                    value="1",
                    in_param_if_snref="bribe",
                ),
            ],
            state_transition_refs=[
                StateTransitionRef(
                    ref_id="charts.annoyed.transitions.get_up",
                    ref_docs=doc_frags,
                    value="false",
                    in_param_if_snref="can_do_backward_flips",
                ),
                StateTransitionRef(
                    ref_id="charts.annoyed.transitions.get_up_to_park",
                    ref_docs=doc_frags,
                    value="true",
                    in_param_if_snref="can_do_backward_flips",
                ),
                StateTransitionRef(
                    ref_id="charts.angry.transitions.get_up",
                    ref_docs=doc_frags,
                ),
            ],
            request_ref=OdxLinkRef.from_id(somersault_requests["start_session"].odx_id),
            semantic="SESSION",
            pos_response_refs=[
                OdxLinkRef.from_id(somersault_positive_responses["session"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["general"].odx_id),
            ],
            functional_class_refs=[
                OdxLinkRef.from_id(somersault_functional_classes["session"].odx_id),
            ],
        ),
    "stop_session":
        DiagService(
            odx_id=OdxLinkId("somersault.service.session_stop", doc_frags),
            short_name="session_stop",
            state_transition_refs=[
                StateTransitionRef(
                    ref_id="charts.angry.transitions.go_to_bed",
                    ref_docs=doc_frags,
                ),
                StateTransitionRef(
                    ref_id="charts.annoyed.transitions.go_to_bed",
                    ref_docs=doc_frags,
                ),
            ],
            semantic="SESSION",
            request_ref=OdxLinkRef.from_id(somersault_requests["stop_session"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(somersault_positive_responses["session"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["general"].odx_id),
            ],
            functional_class_refs=[
                OdxLinkRef.from_id(somersault_functional_classes["session"].odx_id)
            ],
        ),
    "tester_present":
        DiagService(
            odx_id=OdxLinkId("somersault.service.tester_present", doc_frags),
            short_name="tester_present",
            audience=Audience(
                enabled_audience_refs=[
                    OdxLinkRef.from_id(somersault_additional_audiences["attentive_admirer"].odx_id),
                    OdxLinkRef.from_id(somersault_additional_audiences["anyone"].odx_id),
                ],
                is_development_raw=False,
            ),
            semantic="TESTERPRESENT",
            request_ref=OdxLinkRef.from_id(somersault_requests["tester_present"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(somersault_positive_responses["tester_ok"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["tester_nok"].odx_id),
            ],
        ),
    "set_operation_params":
        DiagService(
            odx_id=OdxLinkId("somersault.service.set_operation_params", doc_frags),
            short_name="set_operation_params",
            semantic="FUNCTION",
            request_ref=OdxLinkRef.from_id(somersault_requests["set_operation_params"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(somersault_positive_responses["set_operation_params"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["general"].odx_id),
            ],
        ),
    "forward_flips":
        DiagService(
            odx_id=OdxLinkId("somersault.service.do_forward_flips", doc_frags),
            short_name="do_forward_flips",
            description=Description.from_string("<p>Do a forward flip.</p>"),
            audience=Audience(
                enabled_audience_refs=[
                    OdxLinkRef.from_id(somersault_additional_audiences["attentive_admirer"].odx_id)
                ],
                is_development_raw=False,
            ),
            semantic="FUNCTION",
            request_ref=OdxLinkRef.from_id(somersault_requests["forward_flips"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(
                    somersault_positive_responses["forward_flips_grudgingly_done"].odx_id),
                # TODO: implement handling of multiple responses
                # OdxLinkRef.from_id(somersault_positive_responses["forward_flips_happily_done"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["flips_not_done"].odx_id),
                # TODO (?): implement handling of multiple possible responses
                # OdxLinkRef.from_id(somersault_negative_responses["stumbled"].odx_id),
                # OdxLinkRef.from_id(somersault_negative_responses["too_dizzy"].odx_id),
                # OdxLinkRef.from_id(somersault_negative_responses["not_sober"].odx_id),
            ],
            functional_class_refs=[
                OdxLinkRef.from_id(somersault_functional_classes["flip"].odx_id)
            ],
        ),
    "backward_flips":
        DiagService(
            odx_id=OdxLinkId("somersault.service.do_backward_flips", doc_frags),
            short_name="do_backward_flips",
            audience=Audience(
                enabled_audience_refs=[
                    OdxLinkRef.from_id(somersault_additional_audiences["attentive_admirer"].odx_id)
                ],
                is_development_raw=False,
            ),
            semantic="FUNCTION",
            request_ref=OdxLinkRef.from_id(somersault_requests["backward_flips"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(
                    somersault_positive_responses["backward_flips_grudgingly_done"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["flips_not_done"].odx_id),
            ],
            functional_class_refs=[
                OdxLinkRef.from_id(somersault_functional_classes["flip"].odx_id)
            ],
        ),
    "report_status":
        DiagService(
            odx_id=OdxLinkId("somersault.service.report_status", doc_frags),
            short_name="report_status",
            audience=Audience(
                disabled_audience_refs=[
                    OdxLinkRef.from_id(somersault_additional_audiences["attentive_admirer"].odx_id)
                ],
                is_aftersales_raw=False,
                is_aftermarket_raw=False,
            ),
            semantic="CURRENTDATA",
            request_ref=OdxLinkRef.from_id(somersault_requests["report_status"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(somersault_positive_responses["status_report"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_negative_responses["general"].odx_id),
            ],
        ),
    "schroedinger":
        DiagService(
            odx_id=OdxLinkId("somersault.service.schroedinger", doc_frags),
            short_name="schroedinger",
            request_ref=OdxLinkRef.from_id(somersault_requests["schroedinger"].odx_id),
            semantic="ROUTINE",
        ),
}

somersault_single_ecu_jobs = {
    "compulsory_program":
        SingleEcuJob(
            odx_id=OdxLinkId("somersault.service.compulsory_program", doc_frags),
            short_name="compulsory_program",
            long_name="Compulsory Program",
            description=Description.from_string("<p>Do several fancy moves.</p>"),
            prog_codes=[
                ProgCode(
                    code_file="jobs.py",
                    syntax="PYTHON3",
                    entrypoint="compulsory_program",
                    revision="1.23.4",
                ),
            ],
        )
}

# communication parameters
tester_present_value = somersault_requests["tester_present"].encode()
tester_pr_value = somersault_positive_responses["tester_ok"].encode()
tester_nr_value = somersault_negative_responses["tester_nok"].encode()
somersault_comparam_refs = [
    ###
    # basic parameters
    ###
    # bus speed
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_11898_2_DWCAN.CP_Baudrate", cp_dwcan_doc_frags),
        value="500000",
        protocol_snref="somersault_protocol",
    ),
    # parameters of the CAN diagnostics frames
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_2.CP_UniqueRespIdTable", cp_iso15765_2_doc_frags),
        value=[
            # CP_CanPhysReqFormat
            "normal segmented 11-bit transmit with FC",
            # CP_CanPhysReqId
            "123",
            # CP_CanPhysReqExtAddr
            "0",
            # CP_CanRespUSDTFormat
            "normal segmented 11-bit receive with FC",
            # CP_CanRespUSDTId
            "456",
            # CP_CanRespUSDTExtAddr
            "0",
            # CP_CanRespUUDTFormat
            "normal unsegmented 11-bit receive",
            # CP_CanRespUUDTId
            "4294967295",  # -> -1. this seems to be mandated by the standard. what a hack!
            # CP_CanRespUUDTExtAddr
            "0",
            # CP_ECULayerShortName
            "Somersault",
        ],
        protocol_snref="somersault_protocol",
    ),
    # timeout for responses [us]
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_RC21CompletionTimeout", cp_iso15765_3_doc_frags),
        value="1000000",
        protocol_snref="somersault_protocol",
    ),
    ###
    # "tester present" message handling
    ###
    # expected "tester present" message
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentMessage", cp_iso15765_3_doc_frags),
        value=f"{tester_present_value.hex()}",
        protocol_snref="somersault_protocol",
    ),
    # a response is mandatory
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentReqRsp", cp_iso15765_3_doc_frags),
        value="Response expected",
        protocol_snref="somersault_protocol",
    ),
    # positive response to "tester present"
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentExpPosResp", cp_iso15765_3_doc_frags),
        value=f"{tester_pr_value.hex()}",
        protocol_snref="somersault_protocol",
    ),
    # negative response to "tester present"
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentExpNegResp", cp_iso15765_3_doc_frags),
        value=f"{tester_nr_value.hex()}",
        protocol_snref="somersault_protocol",
    ),
    # when a tester present message must be send
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentSendType", cp_iso15765_3_doc_frags),
        value="On idle",
        protocol_snref="somersault_protocol",
    ),
    # "tester present" messages are send directly to the CAN IDs
    # (i.e., they are not embedded in the ISO-TP telegram?)
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentAddrMode", cp_iso15765_3_doc_frags),
        value="Physical",
        protocol_snref="somersault_protocol",
    ),
    # is a response from the ECU to "tester present" messages expected
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_3.CP_TesterPresentReqRsp", cp_iso15765_3_doc_frags),
        value="Response expected",
        protocol_snref="somersault_protocol",
    ),
    ###
    # ISO-TP parameters:
    ###
    # maximum number of frames between flow control ACKs
    ComparamInstance(
        spec_ref=OdxLinkRef("ISO_15765_2.CP_BlockSize", cp_iso15765_2_doc_frags),
        value="4",
        protocol_snref="somersault_protocol",
    ),
]

somersault_base_diag_data_dictionary_spec = DiagDataDictionarySpec(
    data_object_props=NamedItemList(
        [x for x in somersault_dops.values() if x.short_name != "schroedinger_dop"] +
        [somersault_dops["schroedinger_base"]]),
    unit_spec=UnitSpec(
        unit_groups=NamedItemList(somersault_unit_groups.values()),
        units=NamedItemList(somersault_units.values()),
        physical_dimensions=NamedItemList(somersault_physical_dimensions.values()),
    ),
    tables=NamedItemList(somersault_tables.values()),
    structures=NamedItemList(somersault_structures.values()),
)

somersault_lazy_diag_data_dictionary_spec = DiagDataDictionarySpec(
    data_object_props=NamedItemList([somersault_dops["schroedinger_lazy"]]))

somersault_assiduous_diag_data_dictionary_spec = DiagDataDictionarySpec(
    data_object_props=NamedItemList([somersault_dops["schroedinger_assiduous"]]))

# diagnostics layer for the protocol
somersault_protocol_raw = ProtocolRaw(
    variant_type=DiagLayerType.PROTOCOL,
    odx_id=OdxLinkId("somersault.protocol", doc_frags),
    short_name="somersault_protocol",
    long_name="Somersault protocol info",
    description=Description.from_string(
        "<p>Protocol information of the somersault ECUs &amp; cetera</p>"),
    comparam_spec_ref=OdxLinkRef("CPS_ISO_15765_3_on_ISO_15765_2", [
        OdxDocFragment("ISO_15765_3_on_ISO_15765_2", DocType.COMPARAM_SPEC)
    ]),
    comparam_refs=somersault_comparam_refs,
)
somersault_protocol = Protocol(diag_layer_raw=somersault_protocol_raw)

# diagnostics layer for the base variant
somersault_base_variant_raw = BaseVariantRaw(
    variant_type=DiagLayerType.BASE_VARIANT,
    odx_id=OdxLinkId("somersault", doc_frags),
    short_name="somersault",
    long_name="Somersault base variant",
    description=Description.from_string("<p>Base variant of the somersault ECU &amp; cetera</p>"),
    functional_classes=NamedItemList(somersault_functional_classes.values()),
    diag_data_dictionary_spec=somersault_base_diag_data_dictionary_spec,
    diag_comms_raw=[*somersault_services.values(), *somersault_single_ecu_jobs.values()],
    requests=NamedItemList(somersault_requests.values()),
    positive_responses=NamedItemList(somersault_positive_responses.values()),
    negative_responses=NamedItemList(somersault_negative_responses.values()),
    global_negative_responses=NamedItemList(somersault_global_negative_responses.values()),
    additional_audiences=NamedItemList(somersault_additional_audiences.values()),
    parent_refs=[ParentRef(layer_ref=OdxLinkRef.from_id(somersault_protocol.odx_id))])
somersault_base_variant = BaseVariant(diag_layer_raw=somersault_base_variant_raw)

##################
# Lazy variant of Somersault ECU: this one is lazy and cuts corners
##################

somersault_lazy_ecu_raw = EcuVariantRaw(
    variant_type=DiagLayerType.ECU_VARIANT,
    odx_id=OdxLinkId("somersault_lazy", doc_frags),
    short_name="somersault_lazy",
    long_name="Somersault lazy ECU",
    description=Description.from_string(
        "<p>Sloppy variant of the somersault ECU (lazy &lt; assiduous)</p>"),
    diag_data_dictionary_spec=somersault_lazy_diag_data_dictionary_spec,
    state_charts=NamedItemList([
        StateChart(
            semantic="annoyed",
            odx_id=OdxLinkId("charts.annoyed.chart", doc_frags),
            short_name="annoyed_chart",
            description=Description.from_string(
                "<p>State chart for a day where the ECU is grumpy</p>"),
            states=NamedItemList([
                State(
                    odx_id=OdxLinkId("charts.annoyed.states.in_bed", doc_frags),
                    short_name="in_bed",
                ),
                State(
                    odx_id=OdxLinkId("charts.annoyed.states.on_street", doc_frags),
                    short_name="on_street",
                ),
                State(
                    odx_id=OdxLinkId("charts.annoyed.states.in_park", doc_frags),
                    short_name="in_park",
                ),
                State(
                    odx_id=OdxLinkId("charts.annoyed.states.at_lunch", doc_frags),
                    short_name="at_lunch",
                ),
            ]),
            start_state_snref="in_bed",
            state_transitions=[
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.get_up", doc_frags),
                    short_name="get_up",
                    source_snref="in_bed",
                    target_snref="on_street",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.get_up_to_park", doc_frags),
                    short_name="get_up_to_park",
                    source_snref="in_bed",
                    target_snref="in_park",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.go_to_bed", doc_frags),
                    short_name="go_to_bed",
                    source_snref="on_street",
                    target_snref="in_bed",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.move", doc_frags),
                    short_name="move",
                    source_snref="on_street",
                    target_snref="in_park",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.stumble", doc_frags),
                    short_name="stumble",
                    source_snref="on_street",
                    target_snref="at_lunch",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.annoyed.transitions.caffeinated", doc_frags),
                    short_name="caffeinated",
                    source_snref="at_lunch",
                    target_snref="in_park",
                ),
            ]),
        StateChart(
            semantic="angry",
            odx_id=OdxLinkId("charts.angry.chart", doc_frags),
            short_name="angry_chart",
            description=Description.from_string(
                "<p>State chart for a day where the ECU has a hissy fit</p>"),
            states=NamedItemList([
                State(
                    odx_id=OdxLinkId("charts.angry.states.in_bed", doc_frags),
                    short_name="in_bed",
                ),
                State(
                    odx_id=OdxLinkId("charts.angry.states.on_street", doc_frags),
                    short_name="on_street",
                ),
                State(
                    odx_id=OdxLinkId("charts.angry.states.in_park", doc_frags),
                    short_name="in_park",
                ),
                State(
                    odx_id=OdxLinkId("charts.angry.states.at_lunch", doc_frags),
                    short_name="at_lunch",
                ),
                State(
                    odx_id=OdxLinkId("charts.angry.states.in_hospital", doc_frags),
                    short_name="in_hospital",
                ),
            ]),
            start_state_snref="in_bed",
            state_transitions=[
                StateTransition(
                    odx_id=OdxLinkId("charts.angry.transitions.get_up", doc_frags),
                    short_name="get_up",
                    source_snref="in_bed",
                    target_snref="on_street",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.angry.transitions.go_to_bed", doc_frags),
                    short_name="go_to_bed",
                    source_snref="on_street",
                    target_snref="in_bed",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.angry.transitions.move", doc_frags),
                    short_name="move",
                    source_snref="on_street",
                    target_snref="in_park",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.angry.transitions.stumble", doc_frags),
                    short_name="stumble",
                    source_snref="on_street",
                    target_snref="at_lunch",
                ),
                StateTransition(
                    odx_id=OdxLinkId("charts.angry.transitions.caffeinated", doc_frags),
                    short_name="caffeinated",
                    source_snref="at_lunch",
                    target_snref="in_park",
                ),
            ],
        ),
    ]),
    parent_refs=[
        ParentRef(
            layer_ref=OdxLinkRef.from_id(somersault_base_variant.odx_id),
            # this variant does not do backflips and does not like
            # being told under which conditions it operates.
            not_inherited_diag_comms=[
                somersault_requests["backward_flips"].short_name,
                somersault_requests["set_operation_params"].short_name,
            ],
        )
    ],
)
somersault_lazy_ecu = EcuVariant(diag_layer_raw=somersault_lazy_ecu_raw)

##################
# Assiduous production variant of Somersault ECU: This one works
# harder than it needs to
##################

# the assiduous ECU also does headstands...
somersault_assiduous_requests = {
    "headstand":
        Request(
            odx_id=OdxLinkId("somersault_assiduous.RQ.do_headstand", doc_frags),
            short_name="do_headstand",
            long_name="Do a headstand & whatever else is required to entertain the customer",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(0x3),
                ),
                ValueParameter(
                    short_name="duration",
                    byte_position=1,
                    dop_ref=OdxLinkRef("somersault.DOP.duration", doc_frags),
                ),
            ]),
        ),
}

# positive responses
somersault_assiduous_positive_responses = {
    "headstand_done":
        Response(
            odx_id=OdxLinkId("somersault_assiduous.PR.headstand_done", doc_frags),
            short_name="headstand_done",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(0x1),
                ),
                # TODO (?): non-byte aligned MatchingRequestParameters
                MatchingRequestParameter(
                    short_name="duration",
                    request_byte_position=1,
                    byte_position=1,
                    byte_length=1,
                ),
            ]),
        ),
}

# negative responses
somersault_assiduous_negative_responses = {
    "fell_over":
        Response(
            odx_id=OdxLinkId("somersault_assiduous.NR.fell_over", doc_frags),
            short_name="fell_over",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="sid",
                    diag_coded_type=somersault_diagcodedtypes["uint8"],
                    byte_position=0,
                    coded_value_raw=str(0x20),
                ),
                # TODO (?): non-byte aligned MatchingRequestParameters
                MatchingRequestParameter(
                    short_name="duration",
                    request_byte_position=1,  # somersault_assiduous_requests["headstand"]["duration"].byte_position
                    byte_position=1,
                    byte_length=1,
                ),
            ]),
        ),
}

# services
somersault_assiduous_services = {
    "headstand":
        DiagService(
            odx_id=OdxLinkId("somersault_assiduous.service.headstand", doc_frags),
            short_name="headstand",
            request_ref=OdxLinkRef.from_id(somersault_assiduous_requests["headstand"].odx_id),
            pos_response_refs=[
                OdxLinkRef.from_id(
                    somersault_assiduous_positive_responses["headstand_done"].odx_id),
            ],
            neg_response_refs=[
                OdxLinkRef.from_id(somersault_assiduous_negative_responses["fell_over"].odx_id),
            ],
            audience=Audience(enabled_audience_refs=[
                OdxLinkRef.from_id(somersault_additional_audiences["attentive_admirer"].odx_id)
            ]),
        ),
}

somersault_assiduous_ecu_raw = EcuVariantRaw(
    variant_type=DiagLayerType.ECU_VARIANT,
    odx_id=OdxLinkId("somersault_assiduous", doc_frags),
    short_name="somersault_assiduous",
    long_name="Somersault assiduous ECU",
    description=Description.from_string(
        "<p>Hard-working variant of the somersault ECU (lazy &lt; assiduous)</p>"),
    diag_data_dictionary_spec=somersault_assiduous_diag_data_dictionary_spec,
    diag_comms_raw=list(somersault_assiduous_services.values()),
    requests=NamedItemList(somersault_assiduous_requests.values()),
    positive_responses=NamedItemList(somersault_assiduous_positive_responses.values()),
    negative_responses=NamedItemList(somersault_assiduous_negative_responses.values()),
    parent_refs=[
        ParentRef(
            layer_ref=OdxLinkRef.from_id(somersault_base_variant.odx_id),
            # this variant does everything which the base variant does
            # and more
            not_inherited_dops=["schroedinger"],
        )
    ],
    comparam_refs=somersault_comparam_refs,
)
somersault_assiduous_ecu = EcuVariant(diag_layer_raw=somersault_assiduous_ecu_raw)

##################
# Container with all ECUs
##################

# create a "diagnosis layer container" object
somersault_dlc = DiagLayerContainer(
    odx_id=OdxLinkId("DLC.somersault", doc_frags),
    short_name=dlc_short_name,
    long_name="Collect all saults in the summer",
    description=Description.from_string(
        "<p>This contains ECUs which do somersaults &amp; cetera</p>"),
    admin_data=somersault_admin_data,
    company_datas=NamedItemList([
        somersault_company_datas["suncus"],
        somersault_company_datas["acme"],
    ]),
    base_variants=NamedItemList([somersault_base_variant]),
    ecu_variants=NamedItemList([somersault_lazy_ecu, somersault_assiduous_ecu]),
    protocols=NamedItemList([somersault_protocol]),
)

# read the communication parameters
comparam_subsets = []
odx_cs_dir = pathlib.Path(__file__).parent / "data"
for odx_cs_filename in (
        "ISO_11898_2_DWCAN.odx-cs",
        "ISO_11898_3_DWFTCAN.odx-cs",
        "ISO_15765_2.odx-cs",
        "ISO_15765_3_CPSS.odx-cs",
        "SAE_J2411_SWCAN_CPSS.odx-cs",
):
    odx_cs_root = ElementTree.parse(odx_cs_dir / odx_cs_filename).getroot()
    subset = odx_cs_root.find("COMPARAM-SUBSET")
    if subset is not None:
        comparam_subsets.append(ComparamSubset.from_et(subset, OdxDocContext(ODX_VERSION, [])))

comparam_specs = []
for odx_c_filename in ("UDSOnCAN_CPS.odx-c",):
    odx_c_root = ElementTree.parse(odx_cs_dir / odx_c_filename).getroot()
    subset = odx_c_root.find("COMPARAM-SPEC")
    if subset is not None:
        comparam_specs.append(ComparamSpec.from_et(subset, OdxDocContext(ODX_VERSION, [])))

# create a database object
database = Database()
database.short_name = "somersault_database"
database._diag_layer_containers = NamedItemList([somersault_dlc])
database._comparam_subsets = NamedItemList(comparam_subsets)
database._comparam_specs = NamedItemList(comparam_specs)
database.add_auxiliary_file("jobs.py",
                            BytesIO(b"""
def compulsory_program():
  print("Hello, World")
"""))

# Create ID mapping and resolve references
database.refresh()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from typing_extensions import override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import DataType, ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType


@dataclass(kw_only=True)
class ReservedParameter(Parameter):
    bit_length: int

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "RESERVED"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return False

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ReservedParameter":
        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        bit_length = int(odxrequire(et_element.findtext("BIT-LENGTH")))

        return ReservedParameter(bit_length=bit_length, **kwargs)

    @override
    def get_static_bit_length(self) -> int | None:
        return self.bit_length

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        encode_state.cursor_byte_position += (encode_state.cursor_bit_position + self.bit_length +
                                              7) // 8
        encode_state.cursor_bit_position = 0
        encode_state.emplace_bytes(b'', self.short_name)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        return decode_state.extract_atomic_value(
            bit_length=self.bit_length,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=None,
            is_highlow_byte_order=False)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class Modification:
    change: str
    reason: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Modification":
        change = odxrequire(et_element.findtext("CHANGE"))
        reason = et_element.findtext("REASON")

        return Modification(change=change, reason=reason)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .element import NamedElement
from .odxdoccontext import OdxDocContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class SwVariable(NamedElement):
    origin: str | None = None
    oid: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SwVariable":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        origin = et_element.findtext("ORIGIN")
        oid = et_element.attrib.get("OID")

        return SwVariable(origin=origin, oid=oid, **kwargs)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass

from ..exceptions import odxraise, odxrequire
from ..odxtypes import AtomicOdxType, DataType
from .compuscale import CompuScale
from .limit import Limit


@dataclass(kw_only=True)
class LinearSegment:
    """Helper class to represent a segment of a piecewise-linear interpolation.

    Multiple compu methods (LINEAR, SCALE-LINEAR, TAB-INTP) require
    linear interpolation. This class centralizes the required
    parameters for a single such segment. (The required parameters are
    extracted from the respective compu method's
    COMPU-INTERNAL-TO-PHYS objects. We do it this way because the
    internal-to-phys objects are rather clunky to work with and
    feature a lot of irrelevant information.)

    """
    offset: float
    factor: float
    denominator: float
    internal_lower_limit: Limit | None = None
    internal_upper_limit: Limit | None = None

    inverse_value: int | float  # value used as inverse if factor is 0

    internal_type: DataType
    physical_type: DataType

    @property
    def physical_lower_limit(self) -> Limit | None:
        return self._physical_lower_limit

    @property
    def physical_upper_limit(self) -> Limit | None:
        return self._physical_upper_limit

    @staticmethod
    def from_compu_scale(scale: CompuScale, *, internal_type: DataType,
                         physical_type: DataType) -> "LinearSegment":
        coeffs = odxrequire(scale.compu_rational_coeffs)

        offset = coeffs.numerators[0]
        factor = 0 if len(coeffs.numerators) == 1 else coeffs.numerators[1]

        denominator = 1.0
        if len(coeffs.denominators) > 0:
            denominator = coeffs.denominators[0]

        inverse_value: int | float = 0
        if scale.compu_inverse_value is not None:
            x = odxrequire(scale.compu_inverse_value).value
            if not isinstance(x, (int, float)):
                odxraise(f"Non-numeric COMPU-INVERSE-VALUE specified ({x!r})")
            inverse_value = x

        internal_lower_limit = scale.lower_limit
        internal_upper_limit = scale.upper_limit

        return LinearSegment(
            offset=offset,
            factor=factor,
            denominator=denominator,
            internal_lower_limit=internal_lower_limit,
            internal_upper_limit=internal_upper_limit,
            inverse_value=inverse_value,
            internal_type=internal_type,
            physical_type=physical_type)

    def __post_init__(self) -> None:
        self.__compute_physical_limits()

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> float | int:
        if not isinstance(internal_value, (int, float)):
            odxraise(f"Internal values of linear compumethods must "
                     f"either be int or float (is: {type(internal_value).__name__})")

        result = (self.offset + self.factor * internal_value) / self.denominator

        if self.physical_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
        ]:
            result = round(result)

        return result

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> float | int:
        if not isinstance(physical_value, (int, float)):
            odxraise(f"Physical values of linear compumethods must "
                     f"either be int or float (is: {type(physical_value).__name__})")

        if abs(self.factor) < 1e-10:
            # "If factor = 0 then COMPU-INVERSE-VALUE shall be specified.
            return self.inverse_value

        result = (physical_value * self.denominator - self.offset) / self.factor

        if self.internal_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
        ]:
            result = round(result)

        return result

    def __compute_physical_limits(self) -> None:
        """Computes the physical limits and stores them in the properties
        self._physical_lower_limit and self._physical_upper_limit.
        This method is called by `__post_init__()`.
        """

        def convert_internal_to_physical_limit(internal_limit: Limit | None) -> Limit | None:
            """Helper method to convert a single internal limit
            """
            if internal_limit is None or internal_limit.value_raw is None:
                return None

            internal_value = self.internal_type.from_string(internal_limit.value_raw)
            physical_value = self.convert_internal_to_physical(internal_value)

            result = Limit(
                value_raw=str(physical_value),
                value_type=self.physical_type,
                interval_type=internal_limit.interval_type)

            return result

        self._physical_lower_limit = None
        self._physical_upper_limit = None

        if self.factor >= 0:
            self._physical_lower_limit = convert_internal_to_physical_limit(
                self.internal_lower_limit)
            self._physical_upper_limit = convert_internal_to_physical_limit(
                self.internal_upper_limit)
        else:
            # If the scaling factor is negative, the lower and upper
            # limit are swapped
            self._physical_lower_limit = convert_internal_to_physical_limit(
                self.internal_upper_limit)
            self._physical_upper_limit = convert_internal_to_physical_limit(
                self.internal_lower_limit)

    def physical_applies(self, physical_value: AtomicOdxType) -> bool:
        """Returns True iff the segment is applicable to a given physical value"""
        # Do type checks
        expected_type = self.physical_type.python_type
        if issubclass(expected_type, float):
            if not isinstance(physical_value, (int, float)):
                return False
        else:
            if not isinstance(physical_value, expected_type):
                return False

        if self._physical_lower_limit is not None and \
           not self._physical_lower_limit.complies_to_lower(physical_value):
            return False

        if self._physical_upper_limit is not None and \
           not self._physical_upper_limit.complies_to_upper(physical_value):
            return False

        return True

    def internal_applies(self, internal_value: AtomicOdxType) -> bool:
        """Returns True iff the segment is applicable to a given internal value"""
        # Do type checks
        expected_type = self.internal_type.python_type
        if issubclass(expected_type, float):
            if not isinstance(internal_value, (int, float)):
                return False
        else:
            if not isinstance(internal_value, expected_type):
                return False

        if self.internal_lower_limit is not None and \
           not self.internal_lower_limit.complies_to_lower(internal_value):
            return False

        if self.internal_upper_limit is not None and \
           not self.internal_upper_limit.complies_to_upper(internal_value):
            return False

        return True



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .diaglayers.basevariant import BaseVariant
from .diaglayers.diaglayer import DiagLayer
from .diaglayers.ecushareddata import EcuSharedData
from .diaglayers.ecuvariant import EcuVariant
from .diaglayers.functionalgroup import FunctionalGroup
from .diaglayers.protocol import Protocol
from .nameditemlist import NamedItemList
from .odxcategory import OdxCategory
from .odxdoccontext import OdxDocContext
from .odxlink import DocType, OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .database import Database


@dataclass(kw_only=True)
class DiagLayerContainer(OdxCategory):
    protocols: NamedItemList[Protocol] = field(default_factory=NamedItemList)
    functional_groups: NamedItemList[FunctionalGroup] = field(default_factory=NamedItemList)
    ecu_shared_datas: NamedItemList[EcuSharedData] = field(default_factory=NamedItemList)
    base_variants: NamedItemList[BaseVariant] = field(default_factory=NamedItemList)
    ecu_variants: NamedItemList[EcuVariant] = field(default_factory=NamedItemList)

    @property
    def diag_layers(self) -> NamedItemList[DiagLayer]:
        return self._diag_layers

    @property
    def ecus(self) -> NamedItemList[EcuVariant]:
        """ECU variants defined in the container

        This property is an alias for `.ecu_variants`"""
        return self.ecu_variants

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagLayerContainer":

        cat = OdxCategory.category_from_et(et_element, context, doc_type=DocType.CONTAINER)
        kwargs = dataclass_fields_asdict(cat)

        protocols = NamedItemList([
            Protocol.from_et(dl_element, context)
            for dl_element in et_element.iterfind("PROTOCOLS/PROTOCOL")
        ])
        functional_groups = NamedItemList([
            FunctionalGroup.from_et(dl_element, context)
            for dl_element in et_element.iterfind("FUNCTIONAL-GROUPS/FUNCTIONAL-GROUP")
        ])
        ecu_shared_datas = NamedItemList([
            EcuSharedData.from_et(dl_element, context)
            for dl_element in et_element.iterfind("ECU-SHARED-DATAS/ECU-SHARED-DATA")
        ])
        base_variants = NamedItemList([
            BaseVariant.from_et(dl_element, context)
            for dl_element in et_element.iterfind("BASE-VARIANTS/BASE-VARIANT")
        ])
        ecu_variants = NamedItemList([
            EcuVariant.from_et(dl_element, context)
            for dl_element in et_element.iterfind("ECU-VARIANTS/ECU-VARIANT")
        ])

        return DiagLayerContainer(
            protocols=protocols,
            functional_groups=functional_groups,
            ecu_shared_datas=ecu_shared_datas,
            base_variants=base_variants,
            ecu_variants=ecu_variants,
            **kwargs)

    def __post_init__(self) -> None:
        self._diag_layers = NamedItemList[DiagLayer](chain(
            self.protocols,
            self.functional_groups,
            self.ecu_shared_datas,
            self.base_variants,
            self.ecu_variants,
        ),)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for protocol in self.protocols:
            result.update(protocol._build_odxlinks())
        for functional_group in self.functional_groups:
            result.update(functional_group._build_odxlinks())
        for ecu_shared_data in self.ecu_shared_datas:
            result.update(ecu_shared_data._build_odxlinks())
        for base_variant in self.base_variants:
            result.update(base_variant._build_odxlinks())
        for ecu_variant in self.ecu_variants:
            result.update(ecu_variant._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for protocol in self.protocols:
            protocol._resolve_odxlinks(odxlinks)
        for functional_group in self.functional_groups:
            functional_group._resolve_odxlinks(odxlinks)
        for ecu_shared_data in self.ecu_shared_datas:
            ecu_shared_data._resolve_odxlinks(odxlinks)
        for base_variant in self.base_variants:
            base_variant._resolve_odxlinks(odxlinks)
        for ecu_variant in self.ecu_variants:
            ecu_variant._resolve_odxlinks(odxlinks)

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        super()._finalize_init(database, odxlinks)

        for protocol in self.protocols:
            protocol._finalize_init(database, odxlinks)
        for functional_group in self.functional_groups:
            functional_group._finalize_init(database, odxlinks)
        for ecu_shared_data in self.ecu_shared_datas:
            ecu_shared_data._finalize_init(database, odxlinks)
        for base_variant in self.base_variants:
            base_variant._finalize_init(database, odxlinks)
        for ecu_variant in self.ecu_variants:
            ecu_variant._finalize_init(database, odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

    def __getitem__(self, key: int | str) -> DiagLayer:
        return self.diag_layers[key]



# SPDX-License-Identifier: MIT
import argparse
import importlib
from typing import Any

import odxtools_porsche
import odxtools_porsche.exceptions

from ..version import __version__ as odxtools_version
from .dummy_sub_parser import DummyTool

# import the tool modules which can be loaded. if a tool
# can't be loaded, add a dummy one
tool_modules: list[Any] = []
for tool_name in ["list", "browse", "snoop", "find", "decode", "compare"]:
    try:
        tool_modules.append(importlib.import_module(f".{tool_name}", package="odxtools.cli"))
    except Exception as e:
        tool_modules.append(DummyTool(tool_name, e))


def start_cli() -> None:
    argparser = argparse.ArgumentParser(
        description="\n".join([
            "Utilities to interact with automotive diagnostic descriptions based on the ODX standard.",
            "",
            "Examples:",
            "  For printing all services use:",
            "   odxtools list ./path/to/database.pdx --services",
            "  For browsing the data base and encoding messages use:",
            "   odxtools browse ./path/to/database.pdx",
        ]),
        prog="odxtools",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    argparser.add_argument(
        "--no-strict",
        action="store_true",
        default=False,
        required=False,
        help="Load the dataset in non-strict mode (which is more robust but might lead to undefined behavior)",
    )

    argparser.add_argument(
        "--version", required=False, action="store_true", help="Print the odxtools version")

    subparsers = argparser.add_subparsers(help="Select a subcommand", dest="subparser_name")

    for tool in tool_modules:
        tool.add_subparser(subparsers)

    args = argparser.parse_args()  # deals with the help message handling

    if args.version:
        print(f"odxtools {odxtools_version}")
        exit()
    if args.subparser_name is None:
        argparser.print_usage()
        exit()

    for tool in tool_modules:
        if tool._odxtools_tool_name_ == args.subparser_name:
            orig_strict = odxtools.exceptions.strict_mode
            odxtools.exceptions.strict_mode = not args.no_strict
            try:
                tool.run(args)
            finally:
                odxtools.exceptions.strict_mode = orig_strict
            return



# SPDX-License-Identifier: MIT

import unittest
from argparse import Namespace
from types import ModuleType
from unittest.mock import MagicMock, patch

import odxtools_porsche.cli.compare as compare
import odxtools_porsche.cli.decode as decode
import odxtools_porsche.cli.find as find
import odxtools_porsche.cli.list as list_tool

browse: ModuleType | None
try:
    import odxtools_porsche.cli.browse as browse
except ImportError:
    browse = None


class UtilFunctions:

    @staticmethod
    def run_list_tool(path_to_pdx_file: str = "./examples/somersault.pdx",
                      no_strict: bool = False,
                      ecu_variants: list[str] | None = None,
                      print_neg_responses: bool = False,
                      ecu_services: list[str] | None = None,
                      print_params: bool = False,
                      print_dops: bool = False,
                      print_all: bool = False,
                      dump_database: bool = False) -> None:
        list_args = Namespace(
            pdx_file=path_to_pdx_file,
            no_strict=no_strict,
            variants=ecu_variants,
            global_negative_responses=print_neg_responses,
            services=ecu_services,
            params=print_params,
            dops=print_dops,
            all=print_all,
            dump_database=dump_database)

        list_tool.run(list_args)

    @staticmethod
    def run_decode_tool(
        data: str,
        decode_data: bool = False,
        path_to_pdx_file: str = "./examples/somersault.pdx",
        ecu_variants: list[str] | None = None,
    ) -> None:
        decode_args = Namespace(
            pdx_file=path_to_pdx_file, variants=ecu_variants, data=data, decode=decode_data)

        decode.run(decode_args)

    @staticmethod
    def run_find_tool(service_names: list[str],
                      path_to_pdx_file: str = "./examples/somersault.pdx",
                      ecu_variants: list[str] | None = None,
                      allow_unknown_bit_lengths: bool = False,
                      no_details: bool = False,
                      dump_database: bool = False) -> None:
        find_args = Namespace(
            pdx_file=path_to_pdx_file,
            variants=ecu_variants,
            service_names=service_names,
            relaxed_output=allow_unknown_bit_lengths,
            no_details=no_details,
            dump_database=dump_database)

        find.run(find_args)

    @staticmethod
    def run_compare_tool(path_to_pdx_file: str = "./examples/somersault.pdx",
                         no_strict: bool = False,
                         ecu_variants: list[str] | None = None,
                         database: list[str] | None = None,
                         no_details: bool = True) -> None:

        compare_args = Namespace(
            pdx_file=path_to_pdx_file,
            no_strict=no_strict,
            variants=ecu_variants,
            database=database,
            no_details=no_details)

        compare.run(compare_args)


class TestCommandLineTools(unittest.TestCase):

    def test_list_tool(self) -> None:

        UtilFunctions.run_list_tool()
        UtilFunctions.run_list_tool(no_strict=True)
        UtilFunctions.run_list_tool(ecu_variants=["somersault"])
        UtilFunctions.run_list_tool(print_neg_responses=True)
        UtilFunctions.run_list_tool(print_params=True)
        UtilFunctions.run_list_tool(print_dops=True)
        UtilFunctions.run_list_tool(print_all=True)
        UtilFunctions.run_list_tool(dump_database=True)
        UtilFunctions.run_list_tool(ecu_services=["session_start"])

    def test_decode_tool(self) -> None:

        UtilFunctions.run_decode_tool(data="3E00")
        UtilFunctions.run_decode_tool(data="3e00")
        UtilFunctions.run_decode_tool(data="3E 00", decode_data=True)
        UtilFunctions.run_decode_tool(data="3E 00", ecu_variants=["somersault_lazy"])
        UtilFunctions.run_decode_tool(data="3E 00")

    def test_find_tool(self) -> None:
        UtilFunctions.run_find_tool(service_names=["headstand"])
        UtilFunctions.run_find_tool(service_names=["headstand"], allow_unknown_bit_lengths=True)
        UtilFunctions.run_find_tool(
            service_names=["headstand"], allow_unknown_bit_lengths=True, no_details=True)

    def test_compare_tool(self) -> None:

        UtilFunctions.run_compare_tool()
        UtilFunctions.run_compare_tool(database=["./examples/somersault_modified.pdx"])
        UtilFunctions.run_compare_tool(no_strict=True)
        UtilFunctions.run_compare_tool(
            database=["./examples/somersault_modified.pdx"], no_details=False)
        UtilFunctions.run_compare_tool(
            database=["./examples/somersault_modified.pdx"], ecu_variants=["somersault_lazy"])
        UtilFunctions.run_compare_tool(ecu_variants=[
            "somersault_lazy", "somersault_assiduous", "somersault_young", "somersault_old"
        ])
        UtilFunctions.run_compare_tool(ecu_variants=["somersault_lazy", "somersault_assiduous"])

    @unittest.skipIf(browse is None, "importing the browse tool failed")
    # browse is an interactive tool, so we need to mock a few
    # functions to make PyInquirer reliably bail out
    @patch("sys.stdout.isatty", return_value=False)
    def test_browse_tool(self, pi_prompt: MagicMock) -> None:
        assert browse is not None
        browse_args = Namespace(pdx_file="./examples/somersault.pdx")
        with self.assertRaises(SystemError):
            # browse can only be run interactively
            browse.run(browse_args)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
import unittest
from xml.etree import ElementTree

from packaging.version import Version

from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.exceptions import odxrequire
from odxtools.nameditemlist import NamedItemList
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.physicaldimension import PhysicalDimension
from odxtools.physicaltype import PhysicalType
from odxtools.request import Request
from odxtools.standardlengthtype import StandardLengthType
from odxtools.unit import Unit
from odxtools.unitspec import UnitSpec

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestUnitSpec(unittest.TestCase):

    def test_read_odx(self) -> None:
        expected = UnitSpec(
            physical_dimensions=NamedItemList([
                PhysicalDimension(
                    odx_id=OdxLinkId("ID.metre", doc_frags),
                    short_name="metre",
                    length_exp=1,
                )
            ]),
            units=NamedItemList([
                Unit(
                    odx_id=OdxLinkId("ID.kilometre", doc_frags),
                    short_name="Kilometre",
                    display_name="km",
                    physical_dimension_ref=OdxLinkRef("ID.metre", doc_frags),
                    factor_si_to_unit=1000,
                    offset_si_to_unit=0,
                )
            ]),
        )
        # Define an example ECU job as odx
        sample_unit_spec_odx = f"""
            <UNIT-SPEC>
                <UNITS>
                    <UNIT ID="{expected.units[0].odx_id.local_id}">
                        <SHORT-NAME>{expected.units[0].short_name}</SHORT-NAME>
                        <DISPLAY-NAME>{expected.units[0].display_name}</DISPLAY-NAME>
                        <FACTOR-SI-TO-UNIT>{expected.units[0].factor_si_to_unit}</FACTOR-SI-TO-UNIT>
                        <OFFSET-SI-TO-UNIT>{expected.units[0].offset_si_to_unit}</OFFSET-SI-TO-UNIT>
                        <PHYSICAL-DIMENSION-REF ID-REF="{odxrequire(expected.units[0].physical_dimension_ref).ref_id}" />
                    </UNIT>
                </UNITS>
                <PHYSICAL-DIMENSIONS>
                    <PHYSICAL-DIMENSION ID="{expected.physical_dimensions[0].odx_id.local_id}">
                        <SHORT-NAME>{expected.physical_dimensions[0].short_name}</SHORT-NAME>
                        <LENGTH-EXP>{expected.physical_dimensions[0].length_exp}</LENGTH-EXP>
                    </PHYSICAL-DIMENSION>
                </PHYSICAL-DIMENSIONS>
            </UNIT-SPEC>
        """
        et_element = ElementTree.fromstring(sample_unit_spec_odx)
        spec = UnitSpec.from_et(et_element, OdxDocContext(Version("2.2.0"), doc_frags))
        self.assertEqual(expected.units, spec.units)
        self.assertEqual(expected.physical_dimensions, spec.physical_dimensions)
        self.assertEqual(expected.unit_groups, spec.unit_groups)
        self.assertEqual(expected, spec)

    def test_resolve_odxlinks(self) -> None:
        unit = Unit(
            odx_id=OdxLinkId("unit_time_id", doc_frags),
            short_name="second",
            display_name="s",
            factor_si_to_unit=1,
            offset_si_to_unit=0,
        )
        dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop_id", doc_frags),
            short_name="dop_sn",
            diag_coded_type=dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
            unit_ref=OdxLinkRef.from_id(unit.odx_id),
        )
        dl_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("BV_id", doc_frags),
            short_name="BaseVariant",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop]),
                unit_spec=UnitSpec(units=NamedItemList([unit])),
            ),
            requests=NamedItemList([
                Request(
                    odx_id=OdxLinkId("rq_id", doc_frags),
                    short_name="rq_sn",
                    parameters=NamedItemList([
                        CodedConstParameter(
                            short_name="sid",
                            diag_coded_type=dct,
                            coded_value_raw=str(0x12),
                        ),
                        ValueParameter(
                            short_name="time",
                            dop_ref=OdxLinkRef.from_id(dop.odx_id),
                        ),
                    ]),
                )
            ]),
        )
        dl = EcuVariant(diag_layer_raw=dl_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(dl._build_odxlinks())
        db = Database()
        dl._resolve_odxlinks(odxlinks)
        dl._finalize_init(db, odxlinks)

        param = dl.requests[0].parameters[1]
        assert isinstance(param, ValueParameter)
        _dop = param.dop
        assert isinstance(_dop, DataObjectProperty)
        self.assertEqual(_dop.unit, unit)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from typing import Optional
from xml.etree import ElementTree

from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext


@dataclass(kw_only=True)
class ExternalDoc:
    description: str | None = None
    href: str

    @staticmethod
    def from_et(et_element: ElementTree.Element | None,
                context: OdxDocContext) -> Optional["ExternalDoc"]:
        if et_element is None:
            return None

        description = et_element.text
        href = odxrequire(et_element.get("HREF"))

        return ExternalDoc(description=description, href=href)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .ratfuncsegment import RatFuncSegment


@dataclass(kw_only=True)
class RatFuncCompuMethod(CompuMethod):
    """A compu method using a rational function

    i.e. internal values are converted to physical ones using the
    function `f(x) = (a0 + a1*x + a2*x^2 ...)/(b0 + b0*x^2 ...)` where `f(x)`
    is the physical value and `x` is the internal value. In contrast
    to `ScaleRatFuncCompuMethod`, this compu method only exhibits a
    single segment)

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.5.
    """

    @property
    def int_to_phys_segment(self) -> RatFuncSegment:
        return self._int_to_phys_segment

    @property
    def phys_to_int_segment(self) -> RatFuncSegment | None:
        return self._phys_to_int_segment

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "RatFuncCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return RatFuncCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.RAT_FUNC,
                  "RatFuncCompuMethod must exhibit RAT-FUNC category")

        odxassert(self.physical_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])
        odxassert(self.internal_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])

        if self.compu_internal_to_phys is None:
            odxraise("RAT-FUNC compu methods require COMPU-INTERNAL-TO-PHYS")
            return

        int_to_phys_scales = self.compu_internal_to_phys.compu_scales
        if len(int_to_phys_scales) != 1:
            odxraise("RAT-FUNC compu methods expect exactly one compu scale within "
                     "COMPU-INTERNAL-TO-PHYS")
            return cast(None, RatFuncCompuMethod)

        self._int_to_phys_segment = RatFuncSegment.from_compu_scale(
            int_to_phys_scales[0], value_type=self.physical_type)

        self._phys_to_int_segment = None
        if self.compu_phys_to_internal is not None:
            phys_to_int_scales = self.compu_phys_to_internal.compu_scales
            if len(phys_to_int_scales) != 1:
                odxraise("RAT-FUNC compu methods expect exactly one compu scale within "
                         "COMPU-PHYS-TO-INTERNAL")
                return cast(None, RatFuncCompuMethod)

            self._phys_to_int_segment = RatFuncSegment.from_compu_scale(
                phys_to_int_scales[0], value_type=self.internal_type)

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        if not self._int_to_phys_segment.applies(internal_value):
            odxraise(f"Cannot decode internal value {internal_value!r}", DecodeError)
            return cast(AtomicOdxType, None)

        return self._int_to_phys_segment.convert(internal_value)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        if self._phys_to_int_segment is None or not self._phys_to_int_segment.applies(
                physical_value):
            odxraise(f"Cannot encode physical value {physical_value!r}", EncodeError)
            return cast(AtomicOdxType, None)

        return self._phys_to_int_segment.convert(physical_value)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        return self._phys_to_int_segment is not None and self._phys_to_int_segment.applies(
            physical_value)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        return self._int_to_phys_segment.applies(internal_value)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from ..createanydiagcodedtype import create_any_diag_coded_type_from_et
from ..decodestate import DecodeState
from ..diagcodedtype import DiagCodedType
from ..encodestate import EncodeState
from ..exceptions import DecodeMismatch, EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkId
from ..odxtypes import AtomicOdxType, DataType, ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType


@dataclass(kw_only=True)
class NrcConstParameter(Parameter):
    """A parameter of type NRC-CONST defines a set of values to be
    matched for a negative response object to apply

    The behaviour of NRC-CONST parameters is similar to CODED-CONST
    parameters in that they allow to specify which coding objects
    apply to a binary string, but in contrast to CODED-CONST
    parameters they allow to specify multiple values. Thus, the value
    of a CODED-CONST parameter is usually set using an overlapping
    VALUE parameter. Since NRC-CONST parameters can only be specified
    for negative responses, they can thus be regarded as a multiplexer
    mechanism that is specific to negative responses.

    See ASAM MCD-2 D (ODX), p. 77-79.

    """

    coded_values_raw: list[str] = field(default_factory=list)
    diag_coded_type: DiagCodedType

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "NRC-CONST"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return False

    @property
    def internal_data_type(self) -> DataType:
        return self.diag_coded_type.base_data_type

    @property
    def coded_values(self) -> list[AtomicOdxType]:
        return self._coded_values

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "NrcConstParameter":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        coded_values_raw = [
            odxrequire(x.text) for x in et_element.iterfind("CODED-VALUES/CODED-VALUE")
        ]
        dct_elem = odxrequire(et_element.find("DIAG-CODED-TYPE"))
        diag_coded_type = create_any_diag_coded_type_from_et(dct_elem, context)

        return NrcConstParameter(
            coded_values_raw=coded_values_raw, diag_coded_type=diag_coded_type, **kwargs)

    def __post_init__(self) -> None:
        self._coded_values = [
            self.diag_coded_type.base_data_type.from_string(val) for val in self.coded_values_raw
        ]

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        result.update(self.diag_coded_type._build_odxlinks())

        return result

    @override
    def get_static_bit_length(self) -> int | None:
        return self.diag_coded_type.get_static_bit_length()

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        # NRC-CONST parameters are not encoding any value on its
        # own. instead, it is supposed to overlap with a value
        # parameter.
        if physical_value is not None:
            odxraise("The value of NRC-CONST parameters cannot be set directly!", EncodeError)

        # TODO (?): extract the parameter and check if it is one of
        # the values of self.coded_values. if not, throw an
        # EncodeMismatch exception! This is probably a bad idea
        # because the parameter which determines the value of the
        # NRC-CONST might possibly be specified after the NRC-CONST.

        # move the cursor forward by the size of the parameter
        bit_pos = encode_state.cursor_bit_position
        bit_len = self.diag_coded_type.get_static_bit_length()

        if bit_len is None:
            odxraise("The diag coded type of NRC-CONST parameters must "
                     "exhibit a static size")
            return

        encode_state.cursor_byte_position += (bit_pos + bit_len + 7) // 8
        encode_state.cursor_bit_position = 0

        encode_state.emplace_bytes(b'', self.short_name)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        # Extract coded value
        coded_value = self.diag_coded_type.decode_from_pdu(decode_state)

        # Check if the coded value in the message is correct.
        if coded_value not in self.coded_values:
            raise DecodeMismatch(f"NRC-CONST parameter {self.short_name} does not apply")

        return coded_value

    def get_description_of_valid_values(self) -> str:
        """return a human-understandable description of valid physical values"""
        return f"One of the constant internal values: {self.coded_values}"



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from ..description import Description
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from .compuconst import CompuConst
from .compuinversevalue import CompuInverseValue
from .compurationalcoeffs import CompuRationalCoeffs
from .limit import Limit


@dataclass(kw_only=True)
class CompuScale:
    """A COMPU-SCALE represents one value range of a COMPU-METHOD.
    """

    short_label: str | None = None
    description: Description | None = None
    lower_limit: Limit | None = None
    upper_limit: Limit | None = None
    compu_inverse_value: CompuInverseValue | None = None
    compu_const: CompuConst | None = None
    compu_rational_coeffs: CompuRationalCoeffs | None = None

    # the following two attributes are not specified for COMPU-SCALE
    # tags in the XML, but they are required to do anything useful
    # with compu scales: The domain type is the input set of the
    # function associated with the compu scale object, whilst the
    # range type represents the output set. IOW, for scales contained
    # by the internal-to-physical mapping function, the domain type is
    # the internal and the range type is the physical type of the
    # compu method. (Vice versa for scales specified by the
    # physical-to-internal mapping function.)
    domain_type: DataType
    range_type: DataType

    @staticmethod
    def compuscale_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                           domain_type: DataType, range_type: DataType) -> "CompuScale":
        short_label = et_element.findtext("SHORT-LABEL")
        description = Description.from_et(et_element.find("DESC"), context)

        lower_limit = Limit.limit_from_et(
            et_element.find("LOWER-LIMIT"), context, value_type=domain_type)
        upper_limit = Limit.limit_from_et(
            et_element.find("UPPER-LIMIT"), context, value_type=domain_type)

        compu_inverse_value = None
        if (cive := et_element.find("COMPU-INVERSE-VALUE")) is not None:
            compu_inverse_value = CompuInverseValue.compuvalue_from_et(cive, data_type=domain_type)

        compu_const = None
        if (cce := et_element.find("COMPU-CONST")) is not None:
            compu_const = CompuConst.compuvalue_from_et(cce, data_type=range_type)

        compu_rational_coeffs: CompuRationalCoeffs | None = None
        if (crc_elem := et_element.find("COMPU-RATIONAL-COEFFS")) is not None:
            compu_rational_coeffs = CompuRationalCoeffs.coeffs_from_et(
                crc_elem, context, value_type=range_type)

        return CompuScale(
            short_label=short_label,
            description=description,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            compu_inverse_value=compu_inverse_value,
            compu_const=compu_const,
            compu_rational_coeffs=compu_rational_coeffs,
            domain_type=domain_type,
            range_type=range_type)

    def applies(self, internal_value: AtomicOdxType) -> bool:

        if self.lower_limit is None and self.upper_limit is None:
            # Everything is allowed: No limits have been specified
            return True
        elif self.upper_limit is None:
            # no upper limit has been specified. the spec says that
            # the value specified by the lower limit is the only one
            # which is allowed (cf section 7.3.6.6.1)
            assert self.lower_limit is not None

            return internal_value == self.lower_limit.value
        elif self.lower_limit is None:
            # only the upper limit has been specified. the spec is
            # ambiguous: it only says that if no upper limit is
            # defined, the lower limit shall also be used as the upper
            # limit and a closed interval type ought to be assumed,
            # but it does not say what happens if the lower limit is
            # not defined (which is allowed by the XSD). We thus
            # assume that if only the upper limit is defined, is
            # treated the same way as if only the lower limit is
            # specified.
            assert self.upper_limit is not None

            return internal_value == self.upper_limit.value

        return self.lower_limit.complies_to_lower(internal_value) and \
            self.upper_limit.complies_to_upper(internal_value)



# SPDX-License-Identifier: MIT
from collections import OrderedDict
from itertools import chain
from os import PathLike
from pathlib import Path
from typing import IO, Any, Union
from xml.etree import ElementTree
from zipfile import ZipFile

from packaging.version import Version

from .comparamspec import ComparamSpec
from .comparamsubset import ComparamSubset
from .diaglayercontainer import DiagLayerContainer
from .diaglayers.basevariant import BaseVariant
from .diaglayers.diaglayer import DiagLayer
from .diaglayers.ecushareddata import EcuSharedData
from .diaglayers.ecuvariant import EcuVariant
from .diaglayers.functionalgroup import FunctionalGroup
from .diaglayers.protocol import Protocol
from .exceptions import odxraise, odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext


class Database:
    """This class internalizes the diagnostic database for various ECUs
    described by a collection of ODX files which are usually collated
    into a single PDX file.
    """

    def __init__(self) -> None:
        self.model_version: Version | None = None
        self.auxiliary_files: OrderedDict[str, IO[bytes]] = OrderedDict()

        # create an empty database object
        self._diag_layer_containers = NamedItemList[DiagLayerContainer]()
        self._comparam_subsets = NamedItemList[ComparamSubset]()
        self._comparam_specs = NamedItemList[ComparamSpec]()
        self._short_name = "odx_database"

    def add_pdx_file(self, pdx_file: Union[str, "PathLike[Any]", IO[bytes], ZipFile]) -> None:
        """Add PDX file to database.
        Either pass the path to the file, an IO with the file content or a ZipFile object.
        """
        if isinstance(pdx_file, ZipFile):
            pdx_zip = pdx_file
        else:
            pdx_zip = ZipFile(pdx_file)
        for zip_member in pdx_zip.namelist():
            # The name of ODX files can end with .odx, .odx-d,
            # .odx-c, .odx-cs, .odx-e, .odx-f, .odx-fd, .odx-m,
            # .odx-v .  We could test for all that, or just make
            # sure that the file's suffix starts with .odx
            p = Path(zip_member)
            if p.suffix.lower().startswith(".odx"):
                root = ElementTree.parse(pdx_zip.open(zip_member)).getroot()
                self._process_xml_tree(root)
            elif p.name.lower() == "index.xml":
                root = ElementTree.parse(pdx_zip.open(zip_member)).getroot()
                db_short_name = odxrequire(root.findtext("SHORT-NAME"))
                self.short_name = db_short_name
            else:
                self.add_auxiliary_file(zip_member, pdx_zip.open(zip_member))

    def add_odx_file(self, odx_file_name: Union[str, "PathLike[Any]"]) -> None:
        self._process_xml_tree(ElementTree.parse(odx_file_name).getroot())

    def add_auxiliary_file(self,
                           aux_file_name: Union[str, "PathLike[Any]"],
                           aux_file_obj: IO[bytes] | None = None) -> None:
        if aux_file_obj is None:
            aux_file_obj = open(aux_file_name, "rb")

        self.auxiliary_files[str(aux_file_name)] = aux_file_obj

    def _process_xml_tree(self, root: ElementTree.Element) -> None:
        dlcs: list[DiagLayerContainer] = []
        comparam_subsets: list[ComparamSubset] = []
        comparam_specs: list[ComparamSpec] = []

        # ODX spec version
        model_version = Version(root.attrib.get("MODEL-VERSION", "2.0"))
        if self.model_version is not None and self.model_version != model_version:
            odxraise(f"Different ODX versions used for the same database (ODX {model_version} "
                     f"and ODX {self.model_version}")

        self.model_version = model_version

        dlc = root.find("DIAG-LAYER-CONTAINER")
        if dlc is not None:
            dlcs.append(DiagLayerContainer.from_et(dlc, OdxDocContext(model_version, [])))

        # In ODX 2.0 there was only COMPARAM-SPEC. In ODX 2.2 the
        # content of COMPARAM-SPEC was moved to COMPARAM-SUBSET
        # and COMPARAM-SPEC became a container for PROT-STACKS and
        # a PROT-STACK references a list of COMPARAM-SUBSET
        cp_subset = root.find("COMPARAM-SUBSET")
        if cp_subset is not None:
            comparam_subsets.append(
                ComparamSubset.from_et(cp_subset, OdxDocContext(model_version, [])))

        cp_spec = root.find("COMPARAM-SPEC")
        if cp_spec is not None:
            if model_version < Version("2.2"):
                comparam_subsets.append(
                    ComparamSubset.from_et(cp_spec, OdxDocContext(model_version, [])))
            else:  # odx >= 2.2
                comparam_specs.append(
                    ComparamSpec.from_et(cp_spec, OdxDocContext(model_version, [])))

        self._diag_layer_containers.extend(dlcs)
        self._comparam_subsets.extend(comparam_subsets)
        self._comparam_specs.extend(comparam_specs)

    def refresh(self) -> None:
        # Create wrapper objects
        self._diag_layers = NamedItemList(
            chain(*[dlc.diag_layers for dlc in self.diag_layer_containers]))

        self._ecu_shared_datas = NamedItemList(
            chain(*[dlc.ecu_shared_datas for dlc in self.diag_layer_containers]))
        self._protocols = NamedItemList(
            chain(*[dlc.protocols for dlc in self.diag_layer_containers]))
        self._functional_groups = NamedItemList(
            chain(*[dlc.functional_groups for dlc in self.diag_layer_containers]))
        self._base_variants = NamedItemList(
            chain(*[dlc.base_variants for dlc in self.diag_layer_containers]))
        self._ecu_variants = NamedItemList(
            chain(*[dlc.ecu_variants for dlc in self.diag_layer_containers]))

        # Build odxlinks
        self._odxlinks = OdxLinkDatabase()
        self._odxlinks.update(self._build_odxlinks())

        # Resolve ODXLINK references
        for subset in self.comparam_subsets:
            subset._resolve_odxlinks(self._odxlinks)

        for spec in self.comparam_specs:
            spec._resolve_odxlinks(self._odxlinks)

        for dlc in self.diag_layer_containers:
            dlc._resolve_odxlinks(self._odxlinks)

        # resolve short name references for containers which do not do
        # inheritance (we can call directly call _resolve_snrefs())
        context = SnRefContext()
        context.database = self

        # let the diaglayers sort out the inherited objects
        for subset in self.comparam_subsets:
            subset._finalize_init(self, self._odxlinks)
        for spec in self.comparam_specs:
            spec._finalize_init(self, self._odxlinks)
        for dlc in self.diag_layer_containers:
            dlc._finalize_init(self, self._odxlinks)

        for subset in self.comparam_subsets:
            subset._resolve_snrefs(context)
        for spec in self.comparam_specs:
            spec._resolve_snrefs(context)
        for dlc in self.diag_layer_containers:
            dlc._resolve_snrefs(context)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result: dict[OdxLinkId, Any] = {}

        for subset in self.comparam_subsets:
            result.update(subset._build_odxlinks())

        for spec in self.comparam_specs:
            result.update(spec._build_odxlinks())

        for dlc in self.diag_layer_containers:
            result.update(dlc._build_odxlinks())

        return result

    @property
    def odxlinks(self) -> OdxLinkDatabase:
        """A map from odx_id to object"""
        return self._odxlinks

    @property
    def short_name(self) -> str:
        return self._short_name

    @short_name.setter
    def short_name(self, value: str) -> None:
        self._short_name = value

    @property
    def ecu_shared_datas(self) -> NamedItemList[EcuSharedData]:
        """All ECU shared data layers defined by this database

        ECU shared data layers act as a kind of shared library for
        data that is common to multiple otherwise unrelated ECUs.

        """
        return self._ecu_shared_datas

    @property
    def protocols(self) -> NamedItemList[Protocol]:
        """All protocol layers defined by this database"""
        return self._protocols

    @property
    def functional_groups(self) -> NamedItemList["FunctionalGroup"]:
        """Functional group layers defined in the database"""
        return self._functional_groups

    @property
    def base_variants(self) -> NamedItemList["BaseVariant"]:
        """Base variants defined in the database"""
        return self._base_variants

    @property
    def ecus(self) -> NamedItemList["EcuVariant"]:
        """ECU variants defined in the database

        This property is an alias for `.ecu_variants`"""
        return self._ecu_variants

    @property
    def ecu_variants(self) -> NamedItemList["EcuVariant"]:
        """ECU variants defined in the database"""
        return self._ecu_variants

    @property
    def diag_layers(self) -> NamedItemList["DiagLayer"]:
        """All diagnostic layers defined in the database"""
        return self._diag_layers

    @property
    def diag_layer_containers(self) -> NamedItemList[DiagLayerContainer]:
        return self._diag_layer_containers

    @diag_layer_containers.setter
    def diag_layer_containers(self, value: NamedItemList[DiagLayerContainer]) -> None:
        self._diag_layer_containers = value

    @property
    def comparam_subsets(self) -> NamedItemList[ComparamSubset]:
        return self._comparam_subsets

    @property
    def comparam_specs(self) -> NamedItemList[ComparamSpec]:
        return self._comparam_specs

    def __repr__(self) -> str:
        return f"Database(model_version={self.model_version}, " \
            f"protocols={[x.short_name for x in self.protocols]}, " \
            f"ecus={[x.short_name for x in self.ecus]}, " \
            f"diag_layer_containers={repr(self.diag_layer_containers)}, " \
            f"comparam_subsets={repr(self.comparam_subsets)}, " \
            f"comparam_specs={repr(self.comparam_specs)})"



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .comparamsubset import ComparamSubset
from .element import IdentifiableElement
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class ProtStack(IdentifiableElement):
    # mandatory in ODX 2.2, but non existent in ODX 2.0
    pdu_protocol_type: str
    physical_link_type: str
    comparam_subset_refs: list[OdxLinkRef] = field(default_factory=list)

    @property
    def comparam_subsets(self) -> NamedItemList[ComparamSubset]:
        return self._comparam_subsets

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ProtStack":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        pdu_protocol_type = odxrequire(et_element.findtext("PDU-PROTOCOL-TYPE"))
        physical_link_type = odxrequire(et_element.findtext("PHYSICAL-LINK-TYPE"))
        comparam_subset_refs = [
            OdxLinkRef.from_et(csr_element, context)
            for csr_element in et_element.iterfind("COMPARAM-SUBSET-REFS/"
                                                   "COMPARAM-SUBSET-REF")
        ]

        return ProtStack(
            pdu_protocol_type=pdu_protocol_type,
            physical_link_type=physical_link_type,
            comparam_subset_refs=comparam_subset_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}
        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._comparam_subsets = NamedItemList[ComparamSubset](
            [odxlinks.resolve(x, ComparamSubset) for x in self.comparam_subset_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from enum import Enum


class CompuCategory(Enum):
    IDENTICAL = "IDENTICAL"
    LINEAR = "LINEAR"
    SCALE_LINEAR = "SCALE-LINEAR"
    TEXTTABLE = "TEXTTABLE"
    COMPUCODE = "COMPUCODE"
    TAB_INTP = "TAB-INTP"
    RAT_FUNC = "RAT-FUNC"
    SCALE_RAT_FUNC = "SCALE-RAT-FUNC"



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .admindata import AdminData
from .decodestate import DecodeState
from .element import IdentifiableElement
from .encodestate import EncodeState
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import ParameterValue
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DopBase(IdentifiableElement):
    """Base class for all (simple and complex) data object properties.

    Any class that a parameter can reference via a DOP-REF (Simple
    DOPs, structures, ...) inherits from this class. All DOPs objects
    implement the `Codec` type protocol.

    """

    admin_data: AdminData | None = None
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DopBase":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = None
        if (admin_data_elem := et_element.find("ADMIN-DATA")) is not None:
            admin_data = AdminData.from_et(admin_data_elem, context)

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return DopBase(admin_data=admin_data, sdgs=sdgs, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

    def get_static_bit_length(self) -> int | None:
        return None

    def is_valid_physical_value(self, physical_value: ParameterValue) -> bool:
        """Determine if a phyical value can be handled by the DOP"""
        raise NotImplementedError

    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:
        """Convert the physical value to bytes and emplace them into a PDU."""
        raise NotImplementedError

    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        """Extract the bytes from the PDU and convert them to the physical value."""
        raise NotImplementedError



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .compuscale import CompuScale


@dataclass(kw_only=True)
class TexttableCompuMethod(CompuMethod):
    """Text table compute methods translate numbers to human readable
    textual descriptions.

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.7.

    """

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "TexttableCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return TexttableCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.TEXTTABLE,
                  "TexttableCompuMethod must exhibit TEXTTABLE category")

        # the spec says that the physical data type shall be
        # A_UNICODE2STRING, but we are a bit more lenient and allow
        # any kind of string...
        odxassert(
            self.physical_type
            in [DataType.A_UNICODE2STRING, DataType.A_UTF8STRING, DataType.A_ASCIISTRING],
            "TEXTTABLE must have string type as its physical datatype.")

        if self.compu_internal_to_phys is None:
            odxraise("TEXTTABLE compu methods must exhibit a COMPU-INTERNAL-TO-PHYS subtag.")
            scales = []
        else:
            scales = self.compu_internal_to_phys.compu_scales

        odxassert(
            all(scale.lower_limit is not None or scale.upper_limit is not None for scale in scales),
            "All scales of TEXTTABLE compu methods must provide limits!")

        self._compu_physical_default_value = None
        citp = odxrequire(self.compu_internal_to_phys)
        if (cdv := citp.compu_default_value) is not None:
            self._compu_physical_default_value = self.physical_type.from_string(odxrequire(cdv.vt))

        self._compu_internal_default_value = None
        cpti = self.compu_phys_to_internal
        if cpti is not None and (cdv := cpti.compu_default_value) is not None:
            self._compu_internal_default_value = self.internal_type.from_string(odxrequire(cdv.v))

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        scales = []
        if (citp := self.compu_internal_to_phys) is not None:
            scales = citp.compu_scales
        matching_scales = [
            x for x in scales if x.compu_const is not None and x.compu_const.value == physical_value
        ]

        if len(matching_scales) == 0:
            if self._compu_internal_default_value is None:
                odxraise(f"Texttable could not encode {physical_value!r}.", EncodeError)
                return cast(None, AtomicOdxType)

            return self._compu_internal_default_value
        elif len(matching_scales) > 1:
            odxraise(f"Texttable could not uniquely encode {physical_value!r}.", EncodeError)

        scale = matching_scales[0]
        if scale.compu_inverse_value is not None and (civ :=
                                                      scale.compu_inverse_value.value) is not None:
            return civ
        elif scale.lower_limit is not None and scale.lower_limit._value is not None:
            return scale.lower_limit._value
        elif scale.upper_limit is not None and scale.upper_limit._value is not None:
            return scale.upper_limit._value

        odxraise(f"Texttable compu method could not encode '{physical_value!r}'.", EncodeError)

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        scales = []
        if (citp := self.compu_internal_to_phys) is not None:
            scales = citp.compu_scales
        matching_scales: list[CompuScale] = [x for x in scales if x.applies(internal_value)]

        if len(matching_scales) == 0:
            if self._compu_physical_default_value is None:
                odxraise(f"Texttable could not decode {internal_value!r}.", DecodeError)
                return cast(None, AtomicOdxType)

            return self._compu_physical_default_value

        if len(matching_scales) > 1:
            odxraise(f"Texttable could not uniquely decode {internal_value!r}.", DecodeError)

        scale = matching_scales[0]

        if scale.compu_const is None:
            odxraise(f"Encountered a COMPU-SCALE with no COMPU-CONST.")
            return cast(None, AtomicOdxType)

        if scale.compu_const.value is not None:
            return scale.compu_const.value

        odxraise(f"Texttable compu method could not decode '{internal_value!r}'.", EncodeError)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        if self._compu_physical_default_value is not None:
            return True

        scales = []
        if (cpti := self.compu_internal_to_phys) is not None:
            scales = cpti.compu_scales

        return any(scale.compu_const.value == physical_value
                   for scale in scales
                   if scale.compu_const is not None)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        if self._compu_internal_default_value is not None:
            return True

        scales = []
        if (citp := self.compu_internal_to_phys) is not None:
            scales = citp.compu_scales

        return any(scale.applies(internal_value) for scale in scales)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Library(IdentifiableElement):
    """
    A library defines a shared library used for single ECU jobs etc.

    It this is basically equivalent to ProgCode.
    """

    code_file: str
    encryption: str | None = None
    syntax: str
    revision: str
    entrypoint: str | None = None

    @property
    def code(self) -> bytes:
        return self._code

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Library":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        code_file = odxrequire(et_element.findtext("CODE-FILE"))
        encryption = et_element.findtext("ENCRYPTION")
        syntax = odxrequire(et_element.findtext("SYNTAX"))
        revision = odxrequire(et_element.findtext("REVISION"))
        entrypoint = et_element.findtext("ENTRYPOINT")

        return Library(
            code_file=code_file,
            encryption=encryption,
            syntax=syntax,
            revision=revision,
            entrypoint=entrypoint,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        aux_file = odxrequire(context.database).auxiliary_files.get(self.code_file)

        if aux_file is None:
            odxraise(f"Reference to auxiliary file '{self.code_file}' "
                     f"could not be resolved")
            self._code: bytes = cast(bytes, None)
            return

        self._code = aux_file.read()
        aux_file.seek(0)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .admindata import AdminData
from .audience import Audience
from .diagclasstype import DiagClassType
from .element import IdentifiableElement
from .exceptions import odxraise, odxrequire
from .functionalclass import FunctionalClass
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .odxtypes import odxstr_to_bool
from .preconditionstateref import PreConditionStateRef
from .relateddiagcommref import RelatedDiagCommRef
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .state import State
from .statetransition import StateTransition
from .statetransitionref import StateTransitionRef
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .diaglayers.protocol import Protocol


@dataclass(kw_only=True)
class DiagComm(IdentifiableElement):
    """Representation of a diagnostic communication object.

    Diagnostic communication objects are diagnostic services,
    single-ECU jobs and multi-ECU jobs.

    """

    admin_data: AdminData | None = None
    sdgs: list[SpecialDataGroup] = field(default_factory=list)
    functional_class_refs: list[OdxLinkRef] = field(default_factory=list)
    audience: Audience | None = None
    protocol_snrefs: list[str] = field(default_factory=list)
    related_diag_comm_refs: list[RelatedDiagCommRef] = field(default_factory=list)
    pre_condition_state_refs: list[PreConditionStateRef] = field(default_factory=list)
    state_transition_refs: list[StateTransitionRef] = field(default_factory=list)

    # attributes
    semantic: str | None = None
    diagnostic_class: DiagClassType | None = None
    is_mandatory_raw: bool | None = None
    is_executable_raw: bool | None = None
    is_final_raw: bool | None = None

    @property
    def functional_classes(self) -> NamedItemList[FunctionalClass]:
        return self._functional_classes

    @property
    def protocols(self) -> NamedItemList["Protocol"]:
        return self._protocols

    @property
    def related_diag_comms(self) -> NamedItemList["DiagComm"]:
        return self._related_diag_comms

    @property
    def pre_condition_states(self) -> NamedItemList[State]:
        return self._pre_condition_states

    @property
    def state_transitions(self) -> NamedItemList[StateTransition]:
        return self._state_transitions

    @property
    def is_mandatory(self) -> bool:
        return self.is_mandatory_raw is True

    @property
    def is_executable(self) -> bool:
        return self.is_executable_raw in (None, True)

    @property
    def is_final(self) -> bool:
        return self.is_final_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagComm":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        functional_class_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("FUNCT-CLASS-REFS/FUNCT-CLASS-REF")
        ]

        audience = None
        if (audience_elem := et_element.find("AUDIENCE")) is not None:
            audience = Audience.from_et(audience_elem, context)

        protocol_snrefs = [
            odxrequire(el.get("SHORT-NAME"))
            for el in et_element.iterfind("PROTOCOL-SNREFS/PROTOCOL-SNREF")
        ]

        related_diag_comm_refs = [
            RelatedDiagCommRef.from_et(el, context)
            for el in et_element.iterfind("RELATED-DIAG-COMM-REFS/RELATED-DIAG-COMM-REF")
        ]

        pre_condition_state_refs = [
            PreConditionStateRef.from_et(el, context)
            for el in et_element.iterfind("PRE-CONDITION-STATE-REFS/PRE-CONDITION-STATE-REF")
        ]

        state_transition_refs = [
            StateTransitionRef.from_et(el, context)
            for el in et_element.iterfind("STATE-TRANSITION-REFS/STATE-TRANSITION-REF")
        ]

        semantic = et_element.attrib.get("SEMANTIC")

        diagnostic_class: DiagClassType | None = None
        if (diagnostic_class_str := et_element.attrib.get("DIAGNOSTIC-CLASS")) is not None:
            try:
                diagnostic_class = DiagClassType(diagnostic_class_str)
            except ValueError:
                odxraise(f"Encountered unknown diagnostic class type '{diagnostic_class_str}'")

        is_mandatory_raw = odxstr_to_bool(et_element.attrib.get("IS-MANDATORY"))
        is_executable_raw = odxstr_to_bool(et_element.attrib.get("IS-EXECUTABLE"))
        is_final_raw = odxstr_to_bool(et_element.attrib.get("IS-FINAL"))

        return DiagComm(
            admin_data=admin_data,
            sdgs=sdgs,
            functional_class_refs=functional_class_refs,
            audience=audience,
            protocol_snrefs=protocol_snrefs,
            related_diag_comm_refs=related_diag_comm_refs,
            pre_condition_state_refs=pre_condition_state_refs,
            state_transition_refs=state_transition_refs,
            semantic=semantic,
            diagnostic_class=diagnostic_class,
            is_mandatory_raw=is_mandatory_raw,
            is_executable_raw=is_executable_raw,
            is_final_raw=is_final_raw,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        if self.admin_data is not None:
            result.update(self.admin_data._build_odxlinks())

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        if self.audience is not None:
            result.update(self.audience._build_odxlinks())

        for pc_ref in self.pre_condition_state_refs:
            result.update(pc_ref._build_odxlinks())

        for st_ref in self.state_transition_refs:
            result.update(st_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data:
            self.admin_data._resolve_odxlinks(odxlinks)

        if self.audience:
            self.audience._resolve_odxlinks(odxlinks)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

        for pc_ref in self.pre_condition_state_refs:
            pc_ref._resolve_odxlinks(odxlinks)

        for st_ref in self.state_transition_refs:
            st_ref._resolve_odxlinks(odxlinks)

        self._related_diag_comms = NamedItemList(
            [odxlinks.resolve(dc_ref, DiagComm) for dc_ref in self.related_diag_comm_refs])
        self._functional_classes = NamedItemList(
            [odxlinks.resolve(fc_ref, FunctionalClass) for fc_ref in self.functional_class_refs])
        self._pre_condition_states = NamedItemList(
            [odxlinks.resolve(st_ref, State) for st_ref in self.pre_condition_state_refs])
        self._state_transitions = NamedItemList(
            [odxlinks.resolve(stt_ref, StateTransition) for stt_ref in self.state_transition_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.admin_data:
            self.admin_data._resolve_snrefs(context)

        if self.audience:
            self.audience._resolve_snrefs(context)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

        for pc_ref in self.pre_condition_state_refs:
            pc_ref._resolve_snrefs(context)

        for st_ref in self.state_transition_refs:
            st_ref._resolve_snrefs(context)

        if TYPE_CHECKING:
            diag_layer = odxrequire(context.diag_layer)
            self._protocols = NamedItemList([
                resolve_snref(prot_snref, getattr(diag_layer, "protocols", []), Protocol)
                for prot_snref in self.protocol_snrefs
            ])
        else:
            diag_layer = odxrequire(context.diag_layer)
            self._protocols = NamedItemList([
                resolve_snref(prot_snref, getattr(diag_layer, "protocols", []))
                for prot_snref in self.protocol_snrefs
            ])



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .physicaldimension import PhysicalDimension
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Unit(IdentifiableElement):
    """
    A unit consists of an ID, short name and a display name.

    Additionally, a unit may reference an SI unit (`.physical_dimension`)
    and an offset to that unit (`factor_si_to_unit`, `offset_si_to_unit`).
    The factor and offset are defined such that the following equation holds true:

    UNIT = FACTOR-SI-TO-UNIT * SI-UNIT + OFFSET-SI-TO-UNIT

    For example: 1km = 1000 * 1m + 0

    Examples
    --------

    A minimal unit representing kilometres:

    ```
    Unit(
        odx_id="kilometre",
        short_name="kilometre",
        display_name="km"
    )
    ```

    A unit that also references a physical dimension:

    ```
    Unit(
        odx_id=OdxLinkId("ID.kilometre", doc_frags),
        short_name="Kilometre",
        display_name="km",
        physical_dimension_ref=OdxLinkRef("ID.metre", doc_frags),
        factor_si_to_unit=1000,
        offset_si_to_unit=0
    )
    # where the physical_dimension_ref references, e.g.:
    PhysicalDimension(odx_id=OdxLinkId("ID.metre", doc_frags), short_name="metre", length_exp=1)
    ```
    """
    display_name: str
    factor_si_to_unit: float | None = None
    offset_si_to_unit: float | None = None
    physical_dimension_ref: OdxLinkRef | None = None

    @property
    def physical_dimension(self) -> PhysicalDimension | None:
        return self._physical_dimension

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Unit":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        display_name = odxrequire(et_element.findtext("DISPLAY-NAME"))

        def read_optional_float(element: ElementTree.Element, name: str) -> float | None:
            if (elem_str := element.findtext(name)) is not None:
                return float(elem_str)
            else:
                return None

        factor_si_to_unit = read_optional_float(et_element, "FACTOR-SI-TO-UNIT")
        offset_si_to_unit = read_optional_float(et_element, "OFFSET-SI-TO-UNIT")
        physical_dimension_ref = OdxLinkRef.from_et(
            et_element.find("PHYSICAL-DIMENSION-REF"), context)

        return Unit(
            display_name=display_name,
            factor_si_to_unit=factor_si_to_unit,
            offset_si_to_unit=offset_si_to_unit,
            physical_dimension_ref=physical_dimension_ref,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._physical_dimension: PhysicalDimension | None = None
        if self.physical_dimension_ref:
            self._physical_dimension = odxlinks.resolve(self.physical_dimension_ref,
                                                        PhysicalDimension)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass
